from typing import Dict, List

import numpy as np
import pandas as pd

from simba.utils.data import detect_bouts
from simba.utils.enums import Methods
from simba.utils.errors import ColumnNotFoundError, InvalidFileTypeError
from simba.utils.read_write import get_fn_ext, read_video_info
from simba.utils.warnings import ThirdPartyAnnotationsInvalidFileFormatWarning


def observer_timestamp_corrector(timestamps: List[str]) -> List[str]:
    corrected_ts = []
    for timestamp in timestamps:
        h, m, s = timestamp.split(":", 3)
        missing_fractions = 9 - len(s)
        if missing_fractions == 0:
            corrected_ts.append(timestamp)
        else:
            corrected_ts.append(f'{h}:{m}:{s}.{"0" * missing_fractions}')
    return corrected_ts


def is_new_boris_version(pd_df: pd.DataFrame):
    """
    Check the format of a boris annotation file.

    In the new version, additional column names are present, while
    others have slightly different name. Here, we check for the presence
    of a column name present only in the newer version.

    :return: True if newer version
    """
    return "Media file name" in list(pd_df.columns)


def read_boris_annotation_files(
    data_paths: List[str],
    error_setting: str,
    video_info_df: pd.DataFrame,
    log_setting: bool = False,
) -> Dict[str, pd.DataFrame]:
    MEDIA_FILE_PATH = "Media file path"
    OBSERVATION_ID = "Observation id"
    TIME = "Time"
    BEHAVIOR = "Behavior"
    STATUS = "Status"

    dfs = {}
    for file_cnt, file_path in enumerate(data_paths):
        _, video_name, _ = get_fn_ext(file_path)
        boris_df = pd.read_csv(file_path)
        try:
            if not is_new_boris_version(boris_df):
                expected_headers = [TIME, MEDIA_FILE_PATH, BEHAVIOR, STATUS]
                start_idx = boris_df[boris_df[OBSERVATION_ID] == TIME].index.values
                df = pd.read_csv(file_path, skiprows=range(0, int(start_idx + 1)))[
                    expected_headers
                ]
            else:
                # Adjust column names to newer BORIS annotation format
                MEDIA_FILE_PATH = "Media file name"
                STATUS = "Behavior type"
                expected_headers = [TIME, MEDIA_FILE_PATH, BEHAVIOR, STATUS]
                df = pd.read_csv(file_path)[expected_headers]
            _, video_base_name, _ = get_fn_ext(df.loc[0, MEDIA_FILE_PATH])
            df.drop(MEDIA_FILE_PATH, axis=1, inplace=True)
            df.columns = ["TIME", "BEHAVIOR", "EVENT"]
            df["TIME"] = df["TIME"].astype(float)
            dfs[video_base_name] = df.sort_values(by="TIME")
        except Exception as e:
            print(e)
            if error_setting == Methods.WARNING.value:
                ThirdPartyAnnotationsInvalidFileFormatWarning(
                    annotation_app="BORIS", file_path=file_path, log_status=log_setting
                )
            elif error_setting == Methods.ERROR.value:
                raise InvalidFileTypeError(
                    msg=f"{file_path} is not a valid BORIS file. See the docs for expected file format."
                )
            else:
                pass
    for video_name, video_df in dfs.items():
        _, _, fps = read_video_info(vid_info_df=video_info_df, video_name=video_name)
        video_df["FRAME"] = (video_df["TIME"] * fps).astype(int)
        video_df.drop("TIME", axis=1, inplace=True)
    return dfs


# video_info_df = read_video_info_csv(file_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/video_info.csv')
#
# df = read_boris_annotation_files(data_paths=['/Users/simon/Downloads/FIXED/c_oxt27_190816_122013_s_trimmcropped.csv'],
#                                 error_setting='WARNING',
#                                  log_setting=False,
#                                  video_info_df=video_info_df)


def read_ethovision_files(
    data_paths: List[str],
    error_setting: str,
    video_info_df: pd.DataFrame,
    log_setting: bool = False,
) -> Dict[str, pd.DataFrame]:
    VIDEO_FILE = "Video file"
    HEADER_LINES = "Number of header lines:"
    RECORDING_TIME = "Recording time"
    BEHAVIOR = "Behavior"
    EVENT = "Event"
    POINT_EVENT = "point event"
    STATE_START = "state start"
    STATE_STOP = "state stop"
    START = "START"
    STOP = "STOP"

    EXPECTED_FIELDS = [RECORDING_TIME, BEHAVIOR, EVENT]

    dfs = {}
    data_paths = [x for x in data_paths if "~$" not in x]
    for file_cnt, file_path in enumerate(data_paths):
        _, video_name, _ = get_fn_ext(filepath=file_path)
        print(
            f"Reading ETHOVISION annotation file ({str(file_cnt + 1)} / {str(len(data_paths))}) ..."
        )
        try:
            df = pd.read_excel(file_path, sheet_name=None)
            sheet_name = list(df.keys())[-1]
            df = pd.read_excel(
                file_path, sheet_name=sheet_name, index_col=0, header=None
            )
            video_path = df.loc[VIDEO_FILE].values[0]
            _, video_name, ext = get_fn_ext(video_path)
            header_n = int(df.loc[HEADER_LINES].values[0]) - 2
            df = df.iloc[header_n:].reset_index(drop=True)
            df.columns = list(df.iloc[0])
            df = df.iloc[2:].reset_index(drop=True)[EXPECTED_FIELDS]
            df.columns = ["TIME", "BEHAVIOR", "EVENT"]
            df = df[df["EVENT"] != POINT_EVENT].reset_index(drop=True)
            df["EVENT"] = df["EVENT"].replace({STATE_START: START, STATE_STOP: STOP})
            dfs[video_name] = df

        except Exception as e:
            if error_setting == Methods.WARNING.value:
                ThirdPartyAnnotationsInvalidFileFormatWarning(
                    annotation_app="ETHOVISION",
                    file_path=file_path,
                    log_status=log_setting,
                )
            elif error_setting == Methods.ERROR.value:
                raise InvalidFileTypeError(
                    msg=f"{file_path} is not a valid ETHOVISION file. See the docs for expected file format."
                )
            else:
                pass

    for video_name, video_df in dfs.items():
        _, _, fps = read_video_info(vid_info_df=video_info_df, video_name=video_name)
        video_df["FRAME"] = (video_df["TIME"] * fps).astype(int)
        video_df.drop("TIME", axis=1, inplace=True)

    return dfs


# video_info_df = read_video_info_csv(file_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/video_info.csv')
#
# df = read_ethovision_files(data_paths=['/Users/simon/Desktop/envs/simba_dev/tests/test_data/import_tests/ethovision_data/correct.xlsx'],
#                                 error_setting='WARNING',
#                                  log_setting=False,
#                                  video_info_df=video_info_df)
#


def read_observer_files(
    data_paths: List[str],
    error_setting: str,
    video_info_df: pd.DataFrame,
    log_setting: bool = False,
) -> Dict[str, pd.DataFrame]:
    TIME_FIELD = "Time_Relative_hmsf"
    VIDEO_NAME_FIELD = "Observation"
    BEHAVIOR_FIELD = "Behavior"
    EVENT_TYPE_FIELD = "Event_Type"
    POINT_EVENT = "Point"
    START = "State start"
    STOP = "State stop"
    EXPECTED_FIELDS = [TIME_FIELD, VIDEO_NAME_FIELD, BEHAVIOR_FIELD, EVENT_TYPE_FIELD]

    dfs = {}
    for file_path in data_paths:
        try:
            df = pd.read_excel(
                file_path, sheet_name=None, usecols=EXPECTED_FIELDS
            ).popitem(last=False)[1]
        except KeyError:
            raise ColumnNotFoundError(
                file_name=file_path, column_name=", ".join(EXPECTED_FIELDS)
            )
        try:
            for video_name in df[VIDEO_NAME_FIELD].unique():
                video_df = df[df[VIDEO_NAME_FIELD] == video_name].reset_index(drop=True)
                video_df = video_df[video_df[EVENT_TYPE_FIELD] != POINT_EVENT]
                video_name = video_df[VIDEO_NAME_FIELD].iloc[0]
                video_df.drop(VIDEO_NAME_FIELD, axis=1, inplace=True)
                video_df[TIME_FIELD] = observer_timestamp_corrector(
                    timestamps=list(video_df[TIME_FIELD].astype(str))
                )
                video_df[TIME_FIELD] = pd.to_timedelta(video_df[TIME_FIELD])
                video_df[EVENT_TYPE_FIELD] = video_df[EVENT_TYPE_FIELD].replace(
                    {START: "START", STOP: "STOP"}
                )
                video_df.columns = ["TIME", "BEHAVIOR", "EVENT"]
                if video_name in list(dfs.keys()):
                    dfs[video_name] = pd.concat(
                        [dfs[video_name], video_df], axis=0
                    ).reset_index(drop=True)
                else:
                    dfs[video_name] = video_df

        except Exception as e:
            if error_setting == Methods.WARNING.value:
                ThirdPartyAnnotationsInvalidFileFormatWarning(
                    annotation_app="OBSERVER",
                    file_path=file_path,
                    log_status=log_setting,
                )
            elif error_setting == Methods.ERROR.value:
                raise InvalidFileTypeError(
                    msg=f"{file_path} is not a valid OBSERVER file. See the docs for expected file format."
                )
            else:
                pass

    for video_name, video_df in dfs.items():
        _, _, fps = read_video_info(vid_info_df=video_info_df, video_name=video_name)
        video_df["FRAME"] = video_df["TIME"].dt.total_seconds() * fps
        video_df["FRAME"] = video_df["FRAME"].apply(np.floor).astype(int)
        video_df.drop("TIME", axis=1, inplace=True)

    return dfs


# video_info_df = read_video_info_csv(file_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/video_info.csv')
#
# df = read_observer_files(data_paths=['/Users/simon/Desktop/envs/troubleshooting/Gosia/source/behaviours/Exp_38/03+11WT_20171010-120856_4_no_dupl_no_audio_fps4_grey-simba_crop_frame_n.xlsx'],
#                          error_setting='WARNING',
#                          log_setting=False,
#                         video_info_df=video_info_df)


def read_solomon_files(
    data_paths: List[str],
    error_setting: str,
    video_info_df: pd.DataFrame,
    log_setting: bool = False,
) -> Dict[str, pd.DataFrame]:
    BEHAVIOR = "Behaviour"
    TIME = "Time"
    EXPECTED_COLUMNS = [TIME, BEHAVIOR]

    dfs = {}
    for file_cnt, file_path in enumerate(data_paths):
        _, file_name, _ = get_fn_ext(file_path)
        _, _, fps = read_video_info(vid_info_df=video_info_df, video_name=file_name)
        try:
            df = pd.read_csv(file_path)[EXPECTED_COLUMNS]
            df = df[~df.isnull().any(axis=1)].reset_index(drop=True)
            df["FRAME"] = df[TIME] * fps
            df["FRAME"] = df["FRAME"].apply(np.floor).astype(int)
            video_df = pd.DataFrame()
            for behavior in df[BEHAVIOR].unique():
                behavior_arr = (
                    df["FRAME"][df[BEHAVIOR] == behavior].reset_index(drop=True).values
                )
                new_arr = np.full((np.max(behavior_arr) + 2), 0)
                for i in behavior_arr:
                    new_arr[i] = 1
                bouts = detect_bouts(
                    data_df=pd.DataFrame(new_arr, columns=[behavior]),
                    target_lst=[behavior],
                    fps=1,
                )[["Event", "Start_frame", "End_frame"]].values
                results = []
                for obs in bouts:
                    results.append([obs[0], "START", obs[1]])
                    results.append([obs[0], "STOP", obs[2]])
                video_df = pd.concat(
                    [
                        video_df,
                        pd.DataFrame(
                            results, columns=["BEHAVIOR", "EVENT", "FRAME"]
                        ).sort_values(by=["FRAME"]),
                    ],
                    axis=0,
                )
            dfs[file_name] = video_df.reset_index(drop=True)

        except Exception as e:
            if error_setting == Methods.WARNING.value:
                ThirdPartyAnnotationsInvalidFileFormatWarning(
                    annotation_app="SOLOMON",
                    file_path=file_path,
                    log_status=log_setting,
                )
            elif error_setting == Methods.ERROR.value:
                raise InvalidFileTypeError(
                    msg=f"{file_path} is not a valid SOLOMON file. See the docs for expected file format."
                )
            else:
                pass

    return dfs


# video_info_df = read_video_info_csv(file_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/video_info.csv')
#
# df = read_solomon_files(data_paths=['/Users/simon/Desktop/envs/simba_dev/tests/test_data/solomon_import/solomon_import/Together_1.csv'],
#                          error_setting='WARNING',
#                          log_setting=False,
#                          video_info_df=video_info_df)


def read_bento_files(
    data_paths: List[str],
    error_setting: str,
    video_info_df: pd.DataFrame,
    log_setting: bool = False,
) -> Dict[str, pd.DataFrame]:
    BENTO = "Bento"
    CHANNEL = "Ch1----------"

    dfs = {}
    for file_cnt, file_path in enumerate(data_paths):
        _, video_name, ext = get_fn_ext(filepath=file_path)
        _, _, fps = read_video_info(vid_info_df=video_info_df, video_name=video_name)
        try:
            data_df = pd.read_csv(
                file_path, delim_whitespace=True, index_col=False, low_memory=False
            )
            start_idx = data_df.index[data_df[BENTO] == CHANNEL].values[0]
            sliced_annot = data_df.iloc[start_idx + 1 :]
            clfs = sliced_annot[sliced_annot[BENTO].str.contains(">")]["Bento"].tolist()
            video_events = []
            for clf_name in clfs:
                start_idx = sliced_annot.index[
                    sliced_annot[BENTO] == f"{clf_name}"
                ].values[0]
                clf_df = sliced_annot.loc[start_idx + 2 :, :]
                end_idx = (
                    clf_df.isnull()[clf_df.isnull().any(axis=1)].idxmax(axis=1).index
                )
                if end_idx.values:
                    end_idx = end_idx.values[0]
                else:
                    end_idx = max(clf_df.index + 1)
                clf_df = (
                    clf_df.loc[: end_idx - 1, :]
                    .reset_index(drop=True)
                    .drop("file", axis=1)
                    .astype(float)
                )
                clf_df.columns = ["START", "STOP"]
                clf_df = clf_df * fps
                for obs in clf_df.values:
                    video_events.append([clf_name, "START", obs[0]])
                    video_events.append([clf_name, "STOP", obs[1]])
            video_df = pd.DataFrame(
                video_events, columns=["BEHAVIOR", "EVENT", "FRAME"]
            )
            video_df["FRAME"] = video_df["FRAME"].astype(int)
            video_df["BEHAVIOR"] = video_df["BEHAVIOR"].str[1:]
            dfs[video_name] = video_df
        except Exception as e:
            if error_setting == Methods.WARNING.value:
                ThirdPartyAnnotationsInvalidFileFormatWarning(
                    annotation_app="BENTO", file_path=file_path, log_status=log_setting
                )
            elif error_setting == Methods.ERROR.value:
                raise InvalidFileTypeError(
                    msg=f"{file_path} is not a valid BENTO file. See the docs for expected file format."
                )
            else:
                pass
    return dfs


# video_info_df = read_video_info_csv(file_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/video_info.csv')
#
# df = read_bento_files(data_paths=['/Users/simon/Desktop/envs/simba_dev/tests/test_data/bento_example/Together_1.annot'],
#                          error_setting='WARNING',
#                          log_setting=False,
#                          video_info_df=video_info_df)


def read_deepethogram_files(
    data_paths: List[str], error_setting: str, log_setting: bool = False
) -> Dict[str, pd.DataFrame]:
    BACKGROUND = "background"
    dfs = {}
    for file_cnt, file_path in enumerate(data_paths):
        _, video_name, _ = get_fn_ext(file_path)
        try:
            data_df = pd.read_csv(file_path, index_col=0)
            data_df.drop(BACKGROUND, axis=1, inplace=True)
            bouts = detect_bouts(
                data_df=data_df, target_lst=list(data_df.columns), fps=1
            )[["Event", "Start_frame", "End_frame"]].values
            results = []
            for obs in bouts:
                results.append([obs[0], "START", obs[1]])
                results.append([obs[0], "STOP", obs[2]])
            dfs[video_name] = (
                pd.DataFrame(results, columns=["BEHAVIOR", "EVENT", "FRAME"])
                .sort_values(by=["FRAME"])
                .reset_index(drop=True)
            )
        except Exception as e:
            if error_setting == Methods.WARNING.value:
                ThirdPartyAnnotationsInvalidFileFormatWarning(
                    annotation_app="DEEPETHOGRAM",
                    file_path=file_path,
                    log_status=log_setting,
                )
            elif error_setting == Methods.ERROR.value:
                raise InvalidFileTypeError(
                    msg=f"{file_path} is not a valid BORIS file. See the docs for expected file format."
                )
            else:
                pass

    return dfs


def fix_uneven_start_stop_count(data: pd.DataFrame) -> pd.DataFrame:
    starts = data["FRAME"][data["EVENT"] == "START"].values
    stops = data["FRAME"][data["EVENT"] == "STOP"].values
    if starts.shape[0] < stops.shape[0]:
        sorted_stops = np.sort(stops)
        for start in starts:
            stop_idx = np.argwhere(sorted_stops > start)[0][0]
            sorted_stops = np.delete(sorted_stops, stop_idx)
        for remove_val in sorted_stops:
            remove_idx = np.argwhere(stops == remove_val)[0][0]
            stops = np.delete(stops, remove_idx)

    if stops.shape[0] < starts.shape[0]:
        sorted_starts = np.sort(starts)
        for stop in stops:
            start_idx = np.argwhere(sorted_starts < stop)[-1][0]
            sorted_starts = np.delete(sorted_starts, start_idx)
        for remove_val in sorted_starts:
            remove_idx = np.argwhere(starts == remove_val)[0][0]
            starts = np.delete(starts, remove_idx)

    return pd.DataFrame({"START": starts, "STOP": stops})


def check_stop_events_prior_to_start_events(df: pd.DataFrame) -> List[int]:
    overlaps_idx = []
    for obs_cnt, obs in enumerate(df.values):
        if obs[0] > obs[1]:
            overlaps_idx.append(obs_cnt)
    return overlaps_idx



__author__ = "Simon Nilsson"

import os
import signal
import sys

import cv2

from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import TextOptions
from simba.utils.read_write import get_video_meta_data
from simba.utils.warnings import FrameRangeWarning


def annotation_video_player():
    """Private methods for playing the video that is being annotated in SimBA GUI"""

    def labelling_log_writer(frame_number: int) -> None:
        f.seek(0)
        f.write(str(frame_number))
        f.truncate()
        f.flush()
        os.fsync(f.fileno())

    def print_video_txt(frame_number: int, video_info: dict) -> None:
        current_time = round((frame_number / video_info["fps"]), 2)
        cv2.putText(
            frame,
            f"F~ {frame_number}",
            (
                TextOptions.BORDER_BUFFER_X.value,
                int((video_info["height"] - spacing_scale)),
            ),
            TextOptions.FONT.value,
            font_size,
            TextOptions.COLOR.value,
            TextOptions.TEXT_THICKNESS.value + 1,
        )
        cv2.putText(
            frame,
            f"T~ {current_time}",
            (
                TextOptions.BORDER_BUFFER_X.value,
                int((video_info["height"] - spacing_scale * 2)),
            ),
            TextOptions.FONT.value,
            font_size,
            TextOptions.COLOR.value,
            TextOptions.TEXT_THICKNESS.value + 1,
        )

    video_path = sys.stdin.readline().encode().decode()
    check_file_exist_and_readable(file_path=video_path)
    project_dir = os.path.dirname(os.path.dirname(video_path))
    cap = cv2.VideoCapture(video_path)
    video_meta_data = get_video_meta_data(video_path=video_path)
    space_scale, res_scale, font_scale = 60, 1500, 2
    max_dim = max(video_meta_data["width"], video_meta_data["height"])
    font_size = float(font_scale / (res_scale / max_dim))
    spacing_scale = int(space_scale / (res_scale / max_dim))

    f = open(os.path.join(project_dir, "labelling_info.txt"), "w")
    time_between_frames = int(1000 / video_meta_data["fps"])
    cv2.namedWindow("Video", cv2.WINDOW_NORMAL)

    while True:
        ret, frame = cap.read()
        key = cv2.waitKey(time_between_frames) & 0xFF
        if key == ord("p"):  ### THE VIDEO IS PAUSED
            while True:
                second_key = cv2.waitKey(1)
                current_video_position = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
                f.seek(0)
                f.write(str(current_video_position - 1))
                f.truncate()
                f.flush()
                os.fsync(f.fileno())
                if second_key == ord("t"):  ## BACK UP TWO FRAME
                    if (current_video_position - 2) < 0:
                        FrameRangeWarning(
                            msg=f"FRAME {current_video_position - 2} CANNOT BE SHOWN",
                            source=annotation_video_player.__name__,
                        )
                    else:
                        cap.set(cv2.CAP_PROP_POS_FRAMES, (current_video_position - 2))
                        ret, frame = cap.read()
                        current_video_position = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
                        print_video_txt(
                            frame_number=current_video_position,
                            video_info=video_meta_data,
                        )
                        cv2.imshow("Video", frame)
                        labelling_log_writer(frame_number=current_video_position)
                elif second_key == ord("s"):  ### BACK UP TEN FRAME
                    if (current_video_position - 11) < 0:
                        FrameRangeWarning(
                            msg=f"FRAME {current_video_position - 11} CANNOT BE SHOWN",
                            source=annotation_video_player.__name__,
                        )
                    else:
                        cap.set(cv2.CAP_PROP_POS_FRAMES, (current_video_position - 11))
                        ret, frame = cap.read()
                        current_video_position = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
                        print_video_txt(
                            frame_number=current_video_position,
                            video_info=video_meta_data,
                        )
                        cv2.imshow("Video", frame)
                        labelling_log_writer(frame_number=current_video_position)
                elif second_key == ord("x"):  ### BACK UP 1s
                    if (current_video_position - video_meta_data["fps"]) < 0:
                        FrameRangeWarning(
                            msg=f'FRAME {current_video_position - video_meta_data["fps"]} CANNOT BE SHOWN',
                            source=annotation_video_player.__name__,
                        )
                    else:
                        cap.set(
                            cv2.CAP_PROP_POS_FRAMES,
                            (current_video_position - video_meta_data["fps"]),
                        )
                        ret, frame = cap.read()
                        current_video_position = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
                        print_video_txt(
                            frame_number=current_video_position,
                            video_info=video_meta_data,
                        )
                        cv2.imshow("Video", frame)
                        labelling_log_writer(frame_number=current_video_position)
                elif second_key == ord("w"):  ### FORWARD 1s
                    if (
                        current_video_position + video_meta_data["fps"]
                    ) > video_meta_data["frame_count"]:
                        FrameRangeWarning(
                            msg=f'FRAME {current_video_position + video_meta_data["fps"]} CANNOT BE SHOWN',
                            source=annotation_video_player.__name__,
                        )
                    else:
                        cap.set(
                            cv2.CAP_PROP_POS_FRAMES,
                            (current_video_position + video_meta_data["fps"]),
                        )
                        ret, frame = cap.read()
                        current_video_position = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
                        print_video_txt(
                            frame_number=current_video_position,
                            video_info=video_meta_data,
                        )
                        cv2.imshow("Video", frame)
                        labelling_log_writer(frame_number=current_video_position)
                elif second_key == ord("o"):  ### FORWARD TWO FRAMES
                    if (current_video_position + 1) > video_meta_data["frame_count"]:
                        FrameRangeWarning(
                            msg=f"FRAME {current_video_position + 1} CANNOT BE SHOWN",
                            source=annotation_video_player.__name__,
                        )
                    else:
                        cap.set(cv2.CAP_PROP_POS_FRAMES, (current_video_position + 1))
                        ret, frame = cap.read()
                        current_video_position = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
                        print_video_txt(
                            frame_number=current_video_position,
                            video_info=video_meta_data,
                        )
                        cv2.imshow("Video", frame)
                        labelling_log_writer(frame_number=current_video_position)

                elif second_key == ord("e"):  ### FORWARD TEN FRAMES
                    if (current_video_position + 9) > video_meta_data["frame_count"]:
                        FrameRangeWarning(
                            msg=f"FRAME {current_video_position + 9} CANNOT BE SHOWN",
                            source=annotation_video_player.__name__,
                        )
                    else:
                        cap.set(cv2.CAP_PROP_POS_FRAMES, (current_video_position + 9))
                        ret, frame = cap.read()
                        current_video_position = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
                        print_video_txt(
                            frame_number=current_video_position,
                            video_info=video_meta_data,
                        )
                        cv2.imshow("Video", frame)
                        labelling_log_writer(frame_number=current_video_position)
                elif second_key == ord("p"):
                    break

                if (second_key == ord("q")) | (cv2.getWindowProperty("Video", 1) == -1):
                    cap.release()
                    cv2.destroyAllWindows()
                    path = os.path.join(project_dir, "subprocess.txt")
                    txtFile = open(path)
                    line = txtFile.readline()
                    if second_key == ord("q"):
                        os.kill(int(line), signal.SIGTERM)
                        break
                    else:
                        try:
                            os.kill(int(line), signal.SIGTERM)
                        except OSError:
                            print(
                                "OSError: Cannot save/read latest image file CSV. Please try again"
                            )

        current_video_position = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
        print_video_txt(frame_number=current_video_position, video_info=video_meta_data)
        cv2.imshow("Video", frame)
        if key == ord("q"):
            break
        if cv2.getWindowProperty("Video", 1) == -1:
            break

    cap.release()
    f.close()
    cv2.destroyAllWindows()
    path = os.path.join(project_dir, "subprocess.txt")
    txtFile = open(path)
    line = txtFile.readline()
    os.kill(int(line), signal.SIGTERM)


if __name__ == "__main__":
    annotation_video_player()



__author__ = "Simon Nilsson"

import os

import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.enums import (ConfigKey, Dtypes, Formats, Methods,
                               MLParamKeys, Options, TagNames)
from simba.utils.errors import InvalidInputError, NoDataError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (read_config_entry, read_simba_meta_files,
                                    write_df)


class GridSearchRandomForestClassifier(ConfigReader, TrainModelMixin):
    """
    Grid-searching random forest models from hyperparameter setting and sampling methods
    stored within the `project_folder/configs` directory of a SimBA project.

    .. note::
       Searches you SimBA project `project_folder/configs` directory for files and builds
       one model per config file. `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-7-train-machine-model>`__.

    :param str config_path: path to SimBA project config file in Configparser format

    Example
    ----------
    >>> _ = GridSearchRandomForestClassifier(config_path='MyConfigPath').run()
    """

    def __init__(self, config_path: str):

        ConfigReader.__init__(self, config_path=config_path)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        TrainModelMixin.__init__(self)
        self.model_dir_out = os.path.join(
            read_config_entry(
                self.config,
                ConfigKey.SML_SETTINGS.value,
                ConfigKey.MODEL_DIR.value,
                data_type=Dtypes.STR.value,
            ),
            "validations",
        )
        if not os.path.exists(self.model_dir_out):
            os.makedirs(self.model_dir_out)
        check_if_filepath_list_is_empty(
            filepaths=self.target_file_paths,
            error_msg="Zero annotation files found in project_folder/csv/targets_inserted, cannot create models.",
        )
        if not os.path.exists(self.configs_meta_dir):
            os.makedirs(self.configs_meta_dir)
        self.meta_file_lst = sorted(read_simba_meta_files(self.configs_meta_dir))
        print(
            "Reading in {} annotated files...".format(str(len(self.target_file_paths)))
        )
        self.data_df, self.frm_idx = self.read_all_files_in_folder_mp_futures(
            self.target_file_paths, self.file_type
        )
        self.frm_idx = pd.DataFrame(
            {"VIDEO": list(self.data_df.index), "FRAME_IDX": self.frm_idx}
        )
        self.data_df = self.check_raw_dataset_integrity(
            self.data_df, logs_path=self.logs_path
        )
        self.data_df = self.drop_bp_cords(df=self.data_df)

    def perform_sampling(self, meta_dict: dict):
        if (
            meta_dict[MLParamKeys.TRAIN_TEST_SPLIT_TYPE.value]
            == Methods.SPLIT_TYPE_FRAMES.value
        ):
            self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
                self.x_df, self.y_df, test_size=meta_dict["train_test_size"]
            )
        elif (
            meta_dict[MLParamKeys.TRAIN_TEST_SPLIT_TYPE.value]
            == Methods.SPLIT_TYPE_BOUTS.value
        ):
            self.x_train, self.x_test, self.y_train, self.y_test = (
                self.bout_train_test_splitter(
                    x_df=self.x_df,
                    y_df=self.y_df,
                    test_size=meta_dict["train_test_size"],
                )
            )
        else:
            raise InvalidInputError(
                msg=f"{meta_dict[MLParamKeys.TRAIN_TEST_SPLIT_TYPE.value]} is not recognized as a valid SPLIT TYPE (OPTIONS: FRAMES, BOUTS).",
                source=self.__class__.__name__,
            )
        if (
            meta_dict[MLParamKeys.UNDERSAMPLE_SETTING.value].lower()
            == Methods.RANDOM_UNDERSAMPLE.value
        ):
            self.x_train, self.y_train = self.random_undersampler(
                self.x_train,
                self.y_train,
                meta_dict[MLParamKeys.UNDERSAMPLE_RATIO.value],
            )
        if (
            meta_dict[MLParamKeys.OVERSAMPLE_SETTING.value].lower()
            == Methods.SMOTEENN.value
        ):
            self.x_train, self.y_train = self.smoteen_oversampler(
                self.x_train,
                self.y_train,
                meta_dict[MLParamKeys.OVERSAMPLE_RATIO.value],
            )
        elif (
            meta_dict[MLParamKeys.OVERSAMPLE_SETTING.value].lower()
            == Methods.SMOTE.value
        ):
            self.x_train, self.y_train = self.smote_oversampler(
                self.x_train,
                self.y_train,
                meta_dict[MLParamKeys.OVERSAMPLE_RATIO.value],
            )

        if meta_dict[MLParamKeys.SAVE_TRAIN_TEST_FRM_IDX.value]:
            train_data = self.frm_idx[
                self.frm_idx.index.isin(self.x_train.index)
            ].set_index("VIDEO")
            test_data = self.frm_idx[
                self.frm_idx.index.isin(self.x_test.index)
            ].set_index("VIDEO")
            write_df(
                df=train_data,
                file_type=Formats.CSV.value,
                save_path=os.path.join(
                    self.model_dir_out, f"train_idx_model_{self.config_cnt}.csv"
                ),
            )
            write_df(
                df=test_data,
                file_type=Formats.CSV.value,
                save_path=os.path.join(
                    self.model_dir_out, f"test_idx_{self.config_cnt}.csv"
                ),
            )
            print(
                f"Frame index for train and test set in model {self.config_cnt} saved in {self.model_dir_out} directory..."
            )

    def run(self):
        self.meta_dicts = self.check_validity_of_meta_files(
            data_df=self.data_df, meta_file_paths=self.meta_file_lst
        )
        if len(self.meta_dicts.keys()) == 0:
            raise NoDataError(
                msg="No valid hyper-parameter config files",
                source=self.__class__.__name__,
            )
        for config_cnt, meta_dict in self.meta_dicts.items():
            self.config_cnt = config_cnt
            self.clf_name = meta_dict[MLParamKeys.CLASSIFIER.value]
            print(
                f"Training model {config_cnt+1}/{len(self.meta_dicts.keys())} ({meta_dict[MLParamKeys.CLASSIFIER.value]})..."
            )
            self.class_names = [
                f"Not_{meta_dict[MLParamKeys.CLASSIFIER.value]}",
                meta_dict[MLParamKeys.CLASSIFIER.value],
            ]
            annotation_cols_to_remove = self.read_in_all_model_names_to_remove(
                self.config, self.clf_cnt, meta_dict[MLParamKeys.CLASSIFIER.value]
            )
            self.x_y_df = self.delete_other_annotation_columns(
                self.data_df, annotation_cols_to_remove
            )
            self.x_df, self.y_df = self.split_df_to_x_y(
                self.x_y_df, meta_dict[MLParamKeys.CLASSIFIER.value]
            )
            self.feature_names = self.x_df.columns
            self.check_sampled_dataset_integrity(x_df=self.x_df, y_df=self.y_df)
            self.perform_sampling(meta_dict=meta_dict)
            print(f"MODEL {config_cnt+1} settings")
            self.print_machine_model_information(meta_dict)
            print(f"# {len(self.feature_names)} features.")
            self.rf_clf = RandomForestClassifier(
                n_estimators=meta_dict[MLParamKeys.RF_ESTIMATORS.value],
                max_features=meta_dict[MLParamKeys.RF_MAX_FEATURES.value],
                max_depth=meta_dict[MLParamKeys.RF_MAX_DEPTH.value],
                n_jobs=-1,
                criterion=meta_dict[MLParamKeys.RF_CRITERION.value],
                min_samples_leaf=meta_dict[MLParamKeys.MIN_LEAF.value],
                bootstrap=True,
                verbose=1,
                class_weight=meta_dict[MLParamKeys.CLASS_WEIGHTS.value],
            )

            print(f"Fitting {self.clf_name} model...")
            self.rf_clf = self.clf_fit(
                clf=self.rf_clf, x_df=self.x_train, y_df=self.y_train
            )
            if (
                meta_dict[MLParamKeys.PERMUTATION_IMPORTANCE.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.calc_permutation_importance(
                    self.x_test,
                    self.y_test,
                    self.rf_clf,
                    self.feature_names,
                    self.clf_name,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if (
                meta_dict[MLParamKeys.LEARNING_CURVE.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.calc_learning_curve(
                    self.x_y_df,
                    self.clf_name,
                    meta_dict[MLParamKeys.LEARNING_CURVE_K_SPLITS.value],
                    meta_dict[MLParamKeys.LEARNING_CURVE_DATA_SPLITS.value],
                    meta_dict[MLParamKeys.TT_SIZE.value],
                    self.rf_clf,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if (
                meta_dict[MLParamKeys.PRECISION_RECALL.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.calc_pr_curve(
                    self.rf_clf,
                    self.x_test,
                    self.y_test,
                    self.clf_name,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if (
                meta_dict[MLParamKeys.EX_DECISION_TREE.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.create_example_dt(
                    self.rf_clf,
                    self.clf_name,
                    self.feature_names,
                    self.class_names,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if meta_dict[MLParamKeys.CLF_REPORT.value] in Options.PERFORM_FLAGS.value:
                self.create_clf_report(
                    self.rf_clf,
                    self.x_test,
                    self.y_test,
                    self.class_names,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if (
                meta_dict[MLParamKeys.IMPORTANCE_LOG.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.create_x_importance_log(
                    self.rf_clf,
                    self.feature_names,
                    self.clf_name,
                    self.model_dir_out,
                    save_file_no=config_cnt,
                )
            if (
                meta_dict[MLParamKeys.IMPORTANCE_BAR_CHART.value]
                in Options.PERFORM_FLAGS.value
            ):
                self.create_x_importance_bar_chart(
                    self.rf_clf,
                    self.feature_names,
                    self.clf_name,
                    self.model_dir_out,
                    meta_dict[MLParamKeys.N_FEATURE_IMPORTANCE_BARS.value],
                    save_file_no=config_cnt,
                )
            if MLParamKeys.SHAP_SCORES.value in meta_dict.keys():
                save_n = (
                    meta_dict[MLParamKeys.SHAP_PRESENT.value]
                    + meta_dict[MLParamKeys.SHAP_ABSENT.value]
                )
                shap_multiprocess = False
                if MLParamKeys.SHAP_SAVE_ITERATION.value in meta_dict.keys():
                    try:
                        save_n = int(meta_dict[MLParamKeys.SHAP_SAVE_ITERATION.value])
                    except ValueError:
                        save_n = (
                            meta_dict[MLParamKeys.SHAP_PRESENT.value]
                            + meta_dict[MLParamKeys.SHAP_ABSENT.value]
                        )
                if MLParamKeys.SHAP_MULTIPROCESS.value in meta_dict.keys():
                    shap_multiprocess = meta_dict[MLParamKeys.SHAP_MULTIPROCESS.value]

                if (
                    meta_dict[MLParamKeys.SHAP_SCORES.value]
                    in Options.PERFORM_FLAGS.value
                ):
                    if not shap_multiprocess in Options.PERFORM_FLAGS.value:
                        self.create_shap_log(
                            ini_file_path=self.config_path,
                            rf_clf=self.rf_clf,
                            x_df=self.x_train,
                            y_df=self.y_train,
                            x_names=self.feature_names,
                            clf_name=self.clf_name,
                            cnt_present=meta_dict[MLParamKeys.SHAP_PRESENT.value],
                            cnt_absent=meta_dict[MLParamKeys.SHAP_ABSENT.value],
                            save_path=self.model_dir_out,
                            save_it=save_n,
                            save_file_no=config_cnt,
                        )
                    else:
                        self.create_shap_log_mp(
                            ini_file_path=self.config_path,
                            rf_clf=self.rf_clf,
                            x_df=self.x_train,
                            y_df=self.y_train,
                            x_names=self.feature_names,
                            clf_name=self.clf_name,
                            cnt_present=meta_dict[MLParamKeys.SHAP_PRESENT.value],
                            cnt_absent=meta_dict[MLParamKeys.SHAP_ABSENT.value],
                            save_path=self.model_dir_out,
                            save_file_no=config_cnt,
                        )

            if MLParamKeys.PARTIAL_DEPENDENCY.value in meta_dict.keys():
                if (
                    meta_dict[MLParamKeys.PARTIAL_DEPENDENCY.value]
                    in Options.PERFORM_FLAGS.value
                ):
                    self.partial_dependence_calculator(
                        clf=self.rf_clf,
                        x_df=self.x_train,
                        clf_name=self.clf_name,
                        save_dir=self.model_dir_out,
                    )
            self.create_meta_data_csv_training_multiple_models(
                meta_dict, self.clf_name, self.model_dir_out, save_file_no=config_cnt
            )
            self.save_rf_model(
                self.rf_clf, self.clf_name, self.model_dir_out, save_file_no=config_cnt
            )
            print(
                "Classifier {} saved in models/validations/model_files directory ...".format(
                    str(self.clf_name + "_" + str(config_cnt))
                )
            )
        stdout_success(
            msg="All models and evaluations complete. The models/evaluation files are in models/validations folders",
            source=self.__class__.__name__,
        )


# test = GridSearchRandomForestClassifier(config_path='/Users/simon/Desktop/envs/troubleshooting/testing_enum/project_folder/project_config.ini')
# test.run()


# test = GridSearchRandomForestClassifier(config_path='/Users/simon/Desktop/envs/troubleshooting/sophie/project_folder/project_config.ini')
# test.run()

# test = TrainMultipleModelsFromMeta(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini')
# test.run()

# test = GridSearchRandomForestClassifier(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.run()

# test = TrainMultipleModelsFromMeta(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.run()

# file_paths = glob.glob('/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/targets_inserted' + '/*.csv')
# df = read_all_files_in_folder_mp(file_paths=file_paths, file_type='csv')
# check_dataset_integrity(df=df)

#
# test = GridSearchRandomForestClassifier(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini')
# test.run()



import time

import numpy as np
from numba import jit, prange


@jit(nopython=True)
def count_values_in_range(data: np.array, ranges: np.array):
    results = np.full((data.shape[0], ranges.shape[0]), 0)
    for i in prange(data.shape[0]):
        for j in prange(ranges.shape[0]):
            lower_bound, upper_bound = ranges[j][0], ranges[j][1]
            results[i][j] = data[i][
                np.logical_and(data[i] >= lower_bound, data[i] <= upper_bound)
            ].shape[0]
    return results


@jit(nopython=True)
def count_values_in_range_mp(data: np.array, ranges: np.array):
    results = np.full((data.shape[0], ranges.shape[0]), 0)
    for i in prange(data.shape[0]):
        for j in prange(ranges.shape[0]):
            lower_bound, upper_bound = ranges[j][0], ranges[j][1]

            results[i][j] = data[i][
                np.logical_and(data[i] >= lower_bound, data[i] <= upper_bound)
            ].shape[0]
    return results


ranges = np.array([[0.0, 0.1], [0.0, 0.5], [0.0, 0.75]])
data = np.random.random((1000000, 100))

# start = time.time()
# results = count_values_in_range(data=data, ranges=ranges)
# print(time.time() - start)

start = time.time()
results = count_values_in_range_mp(data=data, ranges=ranges)
print(time.time() - start)



__author__ = "Simon Nilsson"

import itertools
import os
import random
from copy import deepcopy

import pandas as pd

from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Unsupervised
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.printing import SimbaTimer, stdout_success

try:
    from cuml import UMAP

    gpu_flag = True
except ModuleNotFoundError:
    from umap import UMAP


class UmapEmbedder(UnsupervisedMixin):
    """
    Methods for grid-search UMAP model fit and transform.
    Defaults to GPU and cuml.UMAP if GPU available. If GPU unavailable, then umap.UMAP.

    :param data_path: Path holding pickled data-set created by `simba.unsupervised.dataset_creator.DatasetCreator.
    :param save_dir: Empty directory where to save the UMAP results.
    :param hyper_parameters: dict holding UMAP hyperparameters in list format.

    :Example I: Fit.
    >>> hyper_parameters = {'n_neighbors': [10, 2], 'min_distance': [1.0], 'spread': [1.0], 'scaler': 'MIN-MAX', 'variance': 0.25}
    >>> data_path = 'unsupervised/project_folder/logs/unsupervised_data_20230416145821.pickle'
    >>> save_dir = 'unsupervised/dr_models'
    >>> config_path = 'unsupervised/project_folder/project_config.ini'
    >>> embedder = UmapEmbedder(data_path=data_path, save_dir=save_dir)
    >>> embedder.fit(hyper_parameters=hyper_parameters)
    """

    def __init__(self):
        super().__init__()

    def fit(self, data_path: str, save_dir: str, hyper_parameters: dict):
        self.data_path = data_path
        check_file_exist_and_readable(file_path=self.data_path)
        self.data = self.read_pickle(data_path=data_path)
        self.umap_df = deepcopy(self.data[Unsupervised.BOUTS_FEATURES.value]).set_index(
            [
                Unsupervised.VIDEO.value,
                Unsupervised.START_FRAME.value,
                Unsupervised.END_FRAME.value,
            ]
        )
        self.save_dir = save_dir
        self.check_that_directory_is_empty(directory=self.save_dir)
        self.low_var_cols, self.hyper_parameters = None, hyper_parameters
        self.check_umap_hyperparameters(hyper_parameters=hyper_parameters)
        self.search_space = list(
            itertools.product(
                *[
                    hyper_parameters[Unsupervised.N_NEIGHBORS.value],
                    hyper_parameters[Unsupervised.MIN_DISTANCE.value],
                    hyper_parameters[Unsupervised.SPREAD.value],
                ]
            )
        )
        print(f"Building {len(self.search_space)} UMAP model(s)...")
        if hyper_parameters[Unsupervised.VARIANCE.value] > 0:
            self.low_var_cols = self.find_low_variance_fields(
                data=self.umap_df,
                variance=hyper_parameters[Unsupervised.VARIANCE.value],
            )
            self.umap_df = self.drop_fields(data=self.umap_df, fields=self.low_var_cols)
        self.scaler = self.define_scaler(
            scaler_name=hyper_parameters[Unsupervised.SCALER.value]
        )
        self.scaler.fit(self.umap_df)
        self.scaled_umap_data = self.scaler_transform(
            data=self.umap_df, scaler=self.scaler
        )
        self.__create_methods_log()
        self.__fit_umaps()
        self.timer.stop_timer()
        stdout_success(
            msg=f"{len(self.search_space)} models saved in {self.save_dir} directory",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __create_methods_log(self):
        self.methods = {}
        self.methods[Unsupervised.SCALER.value] = self.scaler
        self.methods[Unsupervised.SCALER_TYPE.value] = self.hyper_parameters[
            Unsupervised.SCALER.value
        ]
        self.methods[Unsupervised.SCALED_DATA.value] = self.scaled_umap_data
        self.methods[Unsupervised.VARIANCE.value] = self.hyper_parameters[
            Unsupervised.VARIANCE.value
        ]
        self.methods[Unsupervised.LOW_VARIANCE_FIELDS.value] = self.low_var_cols
        self.methods[Unsupervised.FEATURE_NAMES.value] = self.scaled_umap_data.columns

    def __fit_umaps(self):
        for cnt, h in enumerate(self.search_space):
            self.model_count = cnt
            self.model = {}
            self.model_timer = SimbaTimer()
            self.model_timer.start_timer()
            self.model[Unsupervised.HASHED_NAME.value] = random.sample(
                self.model_names, 1
            )[0]
            self.model[Unsupervised.PARAMETERS.value] = {
                Unsupervised.N_NEIGHBORS.value: h[0],
                Unsupervised.MIN_DISTANCE.value: h[1],
                Unsupervised.SPREAD.value: h[2],
            }
            self.model[Unsupervised.MODEL.value] = UMAP(
                min_dist=self.model[Unsupervised.PARAMETERS.value][
                    Unsupervised.MIN_DISTANCE.value
                ],
                n_neighbors=int(
                    self.model[Unsupervised.PARAMETERS.value][
                        Unsupervised.N_NEIGHBORS.value
                    ]
                ),
                spread=self.model[Unsupervised.PARAMETERS.value][
                    Unsupervised.SPREAD.value
                ],
                metric=Unsupervised.EUCLIDEAN.value,
                verbose=2,
            )
            self.model[Unsupervised.MODEL.value].fit(self.scaled_umap_data.values)
            results = {}
            results[Unsupervised.DATA.value] = self.data
            results[Unsupervised.METHODS.value] = self.methods
            results[Unsupervised.DR_MODEL.value] = self.model
            self.__save(data=results)

    def __save(self, data: dict) -> None:
        self.write_pickle(
            data=data,
            save_path=os.path.join(
                self.save_dir, f"{self.model[Unsupervised.HASHED_NAME.value]}.pickle"
            ),
        )
        self.model_timer.stop_timer()
        stdout_success(
            msg=f"Model {self.model_count+1}/{len(self.search_space)} ({self.model[Unsupervised.HASHED_NAME.value]}) saved...",
            elapsed_time=self.model_timer.elapsed_time,
        )

    def transform(
        self,
        data_path: str,
        model: str or dict,
        settings: dict,
        save_dir: str or None = None,
    ):
        """

        :param data_path:
        :param model:
        :param settings:
        :param save_dir:
        :return:

        :Example I: Transform.
        >>> data_path = 'unsupervised/project_folder/logs/unsupervised_data_20230416145821.pickle'
        >>> save_dir = 'unsupervised/transformed_umap'
        >>> settings = {'DATA': 'RAW', 'format': 'csv'}
        >>> embedder = UmapEmbedder(data_path=data_path, save_dir=save_dir)
        >>> embedder.transform(model='unsupervised/dr_models/boring_lederberg.pickle', settings=settings)

        """

        timer = SimbaTimer(start=True)
        if isinstance(model, str):
            check_file_exist_and_readable(file_path=model)
            model = self.read_pickle(data_path=model)

        check_file_exist_and_readable(file_path=data_path)
        data = self.read_pickle(data_path=data_path)
        self.umap_df = deepcopy(data[Unsupervised.BOUTS_FEATURES.value]).set_index(
            [
                Unsupervised.VIDEO.value,
                Unsupervised.START_FRAME.value,
                Unsupervised.END_FRAME.value,
            ]
        )
        self.umap_df = self.drop_fields(
            data=self.umap_df,
            fields=model[Unsupervised.METHODS.value][
                Unsupervised.LOW_VARIANCE_FIELDS.value
            ],
        )
        self.scaled_umap_data = self.scaler_transform(
            data=self.umap_df,
            scaler=model[Unsupervised.METHODS.value][Unsupervised.SCALER.value],
        )
        self.check_expected_fields(
            data_fields=list(self.scaled_umap_data.columns),
            expected_fields=model[Unsupervised.METHODS.value][
                Unsupervised.FEATURE_NAMES.value
            ],
        )
        self.results = pd.DataFrame(
            model[Unsupervised.DR_MODEL.value][Unsupervised.MODEL.value].transform(
                self.scaled_umap_data
            ),
            columns=["X", "Y"],
            index=self.umap_df.index,
        )
        if settings[Unsupervised.DATA.value] == Unsupervised.SCALED.value:
            self.results = pd.concat([self.scaled_umap_data, self.results], axis=1)
        elif settings[Unsupervised.DATA.value] == Unsupervised.RAW.value:
            self.results = pd.concat([self.umap_df, self.results], axis=1)
        if save_dir:
            save_path = os.path.join(
                save_dir,
                f"Transformed_{model[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
            )
            if settings[Unsupervised.FORMAT.value] is Unsupervised.CSV.value:
                self.results.to_csv(save_path)
            timer.stop_timer()
            print(
                f"Transformed data saved at {save_dir} (elapsed time: {timer.elapsed_time_str}s)"
            )


# data_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/logs/unsupervised_data_20230416145821.pickle'
# save_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/transformed_umap'
# settings = {'DATA': 'RAW', 'format': 'csv'}
# embedder = UmapEmbedder(data_path=data_path, save_dir=save_dir)
# embedder.transform(model='/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models/boring_lederberg.pickle', settings=settings)


# model_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models/funny_heisenberg.pickle'
# data_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/logs/unsupervised_data_20230222150701.pickle'
# save_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models/'
# _ = UMAPTransform(model_path=model_path, data_path=data_path, save_dir=save_dir, settings=settings)

#
#
#
#
#

# hyper_parameters = {'n_neighbors': [10, 2], 'min_distance': [1.0], 'spread': [1.0], 'scaler': 'MIN-MAX', 'variance': 0.25}
# data_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/logs/unsupervised_data_20230416145821.pickle'
# save_dir = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/dr_models'
# config_path = '/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini'
# embedder = UmapEmbedder(data_path=data_path, save_dir=save_dir)
# embedder.fit(hyper_parameters=hyper_parameters)



__author__ = "Simon Nilsson"

import os
from copy import deepcopy

import numpy as np
import pandas as pd

from simba.data_processors.interpolation_smoothing import Interpolate, Smooth
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.enums import Formats, Methods
from simba.utils.errors import BodypartColumnNotFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_all_videos_in_project,
                                    get_video_meta_data, write_df)


class MADLCImporterH5(ConfigReader, PoseImporterMixin):
    """
    Importing multi-animal deeplabcut (maDLC) pose-estimation data (in H5 format)
    into a SimBA project in parquet or CSV format.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str data_folder: Path to folder containing maDLC data in ``.h5`` format.
    :parameter str file_type: Method used to perform pose-estimation in maDLC. OPTIONS: `skeleton`, `box`, `ellipse`.
    :param List[str] id_lst: Names of animals.
    :parameter str interpolation_setting: String defining the pose-estimation interpolation method. OPTIONS: 'None', 'Animal(s): Nearest',
        'Animal(s): Linear', 'Animal(s): Quadratic','Body-parts: Nearest', 'Body-parts: Linear',
        'Body-parts: Quadratic'.
    :parameter dict smoothing_settings: Dictionary defining the pose estimation smoothing method. EXAMPLE: {'Method': 'Savitzky Golay',
        'Parameters': {'Time_window': '200'}})

    .. note::
       `Multi-animal import tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Multi_animal_pose.md>`__.

    Examples
    -----
    >>> _ = MADLCImporterH5(config_path=r'MyConfigPath', data_folder=r'maDLCDataFolder', file_type='ellipse', id_lst=['Animal_1', 'Animal_2'], interpolation_settings='None', smoothing_settings={'Method': 'None', 'Parameters': {'Time_window': '200'}}).run()

    References
    ----------
    .. [1] Lauer et al., Multi-animal pose estimation, identification and tracking with DeepLabCut, `Nature Methods`,
           2022.
    """

    def __init__(
        self,
        config_path: str,
        data_folder: str,
        file_type: str,
        id_lst: list,
        interpolation_settings: str,
        smoothing_settings: dict,
    ):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PoseImporterMixin.__init__(self)

        self.interpolation_settings, self.smoothing_settings = (
            interpolation_settings,
            smoothing_settings,
        )
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(
            self.logs_path, f"data_import_log_{self.datetime}.csv"
        )
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
        self.input_data_paths = self.find_data_files(
            dir=self.data_folder, extensions=Formats.DLC_FILETYPES.value[file_type]
        )
        self.data_and_videos_lk = self.link_video_paths_to_data_paths(
            data_paths=self.input_data_paths,
            video_paths=self.video_paths,
            str_splits=Formats.DLC_NETWORK_FILE_NAMES.value,
        )
        if self.pose_setting is Methods.USER_DEFINED.value:
            self.__update_config_animal_cnt()
        if self.animal_cnt > 1:
            self.check_multi_animal_status()
            self.animal_bp_dict = self.create_body_part_dictionary(
                self.multi_animal_status,
                self.id_lst,
                self.animal_cnt,
                self.x_cols,
                self.y_cols,
                self.p_cols,
                self.clr_lst,
            )
            self.update_bp_headers_file()
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} file(s)...")

    def run(self):
        import_log = pd.DataFrame(
            columns=[
                "VIDEO",
                "IMPORT_TIME",
                "IMPORT_SOURCE",
                "INTERPOLATION_SETTING",
                "SMOOTHING_SETTING",
            ]
        )
        for cnt, (video_name, video_data) in enumerate(self.data_and_videos_lk.items()):
            video_timer = SimbaTimer(start=True)
            self.add_spacer, self.frame_no, self.video_data, self.video_name = (
                2,
                1,
                video_data,
                video_name,
            )
            print(f"Processing {video_name} ...")
            self.data_df = (
                pd.read_hdf(video_data["DATA"])
                .replace([np.inf, -np.inf], np.nan)
                .fillna(0)
            )
            if len(self.data_df.columns) != len(self.bp_headers):
                raise BodypartColumnNotFoundError(
                    msg=f'The number of body-parts in data file {video_data["DATA"]} do not match the number of body-parts in your SimBA project. '
                    f"The number of of body-parts expected by your SimBA project is {int(len(self.bp_headers) / 3)}. "
                    f'The number of of body-parts contained in data file {video_data["DATA"]} is {int(len(self.data_df.columns) / 3)}. '
                    f"Make sure you have specified the correct number of animals and body-parts in your project."
                )
            self.data_df.columns = self.bp_headers
            self.out_df = deepcopy(self.data_df)
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(
                    animal_bp_dict=self.animal_bp_dict,
                    video_info=get_video_meta_data(video_data["VIDEO"]),
                    data_df=self.data_df,
                    video_path=video_data["VIDEO"],
                )
                self.multianimal_identification()
            self.save_path = os.path.join(
                os.path.join(self.input_csv_dir, f"{self.video_name}.{self.file_type}")
            )
            write_df(
                df=self.out_df,
                file_type=self.file_type,
                save_path=self.save_path,
                multi_idx_header=True,
            )
            if self.interpolation_settings != "None":
                self.__run_interpolation()
            if self.smoothing_settings["Method"] != "None":
                self.__run_smoothing()
            video_timer.stop_timer()
            stdout_success(
                msg=f"Video {video_name} data imported...",
                elapsed_time=video_timer.elapsed_time_str,
            )
        self.timer.stop_timer()
        stdout_success(
            msg="All maDLC H5 data files imported",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __run_interpolation(self):
        print(
            "Interpolating missing values in video {} (Method: {}) ...".format(
                self.video_name, self.interpolation_settings
            )
        )
        _ = Interpolate(
            input_path=self.save_path,
            config_path=self.config_path,
            method=self.interpolation_settings,
            initial_import_multi_index=True,
        )

    def __run_smoothing(self):
        print(
            f'Performing {self.smoothing_settings["Method"]} smoothing on video {self.video_name}...'
        )
        Smooth(
            config_path=self.config_path,
            input_path=self.save_path,
            time_window=int(self.smoothing_settings["Parameters"]["Time_window"]),
            smoothing_method=self.smoothing_settings["Method"],
            initial_import_multi_index=True,
        )


# test = MADLCImporterH5(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/h5',
#                    file_type='ellipse',
#                    id_lst=['Simon', 'JJ'],
#                    interpolation_settings='Body-parts: Nearest',
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.run()



import functools
import glob
import multiprocessing
import os
import platform
from copy import deepcopy
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import cv2
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.enums import OS, Formats, TagNames
from simba.utils.errors import CountError, InvalidFilepathError
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    get_video_meta_data, read_df)
from simba.utils.warnings import FrameRangeWarning


def pose_plotter_mp(
    data: pd.DataFrame,
    video_meta_data: dict,
    video_path: str,
    bp_dict: dict,
    circle_size: int,
    video_save_dir: Union[str, os.PathLike],
):
    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    group_cnt = int(data.iloc[0]["group"])
    data = data.drop(["group"], axis=1)
    start_frm, current_frm, end_frm = data.index[0], data.index[0], data.index[-1]
    save_path = os.path.join(video_save_dir, f"{group_cnt}.mp4")
    writer = cv2.VideoWriter(
        save_path,
        fourcc,
        video_meta_data["fps"],
        (video_meta_data["width"], video_meta_data["height"]),
    )
    cap = cv2.VideoCapture(video_path)
    cap.set(1, start_frm)

    while current_frm < end_frm:
        ret, img = cap.read()
        if ret:
            for animal_name, animal_data in bp_dict.items():
                for x_name, y_name in zip(animal_data["X_bps"], animal_data["Y_bps"]):
                    bp_tuple = (
                        int(data.at[current_frm, x_name]),
                        int(data.at[current_frm, y_name]),
                    )
                    cv2.circle(img, bp_tuple, circle_size, animal_data["colors"], -1)
            writer.write(img)
            current_frm += 1
            print(f"Multi-processing video frame {current_frm} on core {group_cnt}...")

    cap.release()
    writer.release()


class PosePlotter:
    """
    Create pose-estimation visualizations from data within a SimBA project folder.

    :param str in_directory: Path to SimBA project directory containing pose-estimation data in parquet or CSV format.
    :param str out_directory: Directory to where to save the pose-estimation videos.
    :param int Size of the circles denoting the location of the pose-estimated body-parts.
    :param Optional[dict] clr_attr: Python dict where animals are keys and color attributes values. E.g., {'Animal_1':  (255, 107, 198)}. If None,
                                    random palettes will be used.

    .. image:: _static/img/pose_plotter.png
       :width: 600
       :align: center

    :example:
    >>> test = PosePlotter(in_dir='project_folder/csv/input_csv', out_dir='/project_folder/test_viz', circle_size=10, core_cnt=1, color_settings={'Animal_1':  'Green', 'Animal_2':  'Red'})
    >>> test.run()
    """

    def __init__(
        self,
        in_dir: Union[str, os.PathLike],
        out_dir: Union[str, os.PathLike],
        circle_size: int,
        core_cnt: int,
        color_settings: Optional[Dict[str, str]] = None,
        sample_time: Optional[int] = None,
    ) -> None:
        config_path, self.data, self.color_settings = (
            os.path.join(Path(in_dir).parents[1], "project_config.ini"),
            {},
            color_settings,
        )
        self.circle_size, self.core_cnt, self.out_dir, self.sample_time = (
            circle_size,
            core_cnt,
            out_dir,
            sample_time,
        )
        self.clrs = get_color_dict()
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)
        if not os.path.isfile(config_path):
            raise InvalidFilepathError(
                msg=f'When visualizing pose-estimation, select an input sub-directory of the project_folder/csv folder. {in_dir} is not a sub-directory to the "project_folder/csv" directory and therefore SimBA cant locate the project_config.ini (expected at {config_path}',
                source=self.__class__.__name__,
            )
        # if platform.system() == OS.MAC.value:
        #     multiprocessing.set_start_method('spawn', force=True)
        self.config = ConfigReader(config_path=config_path, read_video_info=False)
        files_found = glob.glob(in_dir + f"/*.{self.config.file_type}")
        check_if_filepath_list_is_empty(
            filepaths=files_found,
            error_msg=f"0 files found in {in_dir} in {self.config.file_type} file format",
        )

        for file in files_found:
            print(self.config.video_dir)
            self.data[file] = self.config.find_video_of_file(
                video_dir=self.config.video_dir, filename=get_fn_ext(file)[1]
            )
        self.original_bp_dict = deepcopy(self.config.animal_bp_dict)

    def run(self):
        for file_cnt, (pose_path, video_path) in enumerate(self.data.items()):
            video_timer = SimbaTimer(start=True)
            video_name = get_fn_ext(pose_path)[1]
            self.temp_folder = os.path.join(self.out_dir, video_name, "temp")
            if os.path.exists(self.temp_folder):
                self.config.remove_a_folder(self.temp_folder)
            os.makedirs(self.temp_folder)
            save_video_path = os.path.join(self.out_dir, f"{video_name}.mp4")
            pose_df = read_df(
                file_path=pose_path,
                file_type=self.config.file_type,
                check_multiindex=True,
            )
            video_meta_data = get_video_meta_data(video_path=video_path)
            if (self.sample_time is None) and (
                video_meta_data["frame_count"] != len(pose_df)
            ):
                FrameRangeWarning(
                    msg=f'The video {video_name} has pose-estimation data for {len(pose_df)} frames, but the video has {video_meta_data["frame_count"]} frames. Ensure the data and video has an equal number of frames.',
                    source=self.__class__.__name__,
                )
            if type(self.sample_time) is int:
                sample_frm_cnt = int(video_meta_data["fps"] * self.sample_time)
                if sample_frm_cnt > len(pose_df):
                    sample_frm_cnt = len(pose_df)
                pose_df = pose_df.iloc[0:sample_frm_cnt]
            pose_df = self.config.insert_column_headers_for_outlier_correction(
                data_df=pose_df, new_headers=self.config.bp_headers, filepath=pose_path
            )
            pose_df = (
                pose_df.apply(pd.to_numeric, errors="coerce")
                .fillna(0)
                .reset_index(drop=True)
            )
            if self.color_settings:
                for cnt, animal in enumerate(self.config.animal_bp_dict.keys()):
                    self.config.animal_bp_dict[animal]["colors"] = self.clrs[
                        self.color_settings[f"Animal_{cnt+1}"]
                    ]
            else:
                for cnt, animal in enumerate(self.config.animal_bp_dict.keys()):
                    self.config.animal_bp_dict[animal]["colors"] = (
                        self.original_bp_dict[animal]["colors"][0]
                    )

            pose_lst, obs_per_split = PlottingMixin().split_and_group_df(
                df=pose_df, splits=self.core_cnt
            )
            print(
                f"Creating pose videos, multiprocessing (chunksize: {self.config.multiprocess_chunksize}, cores: {self.core_cnt})..."
            )
            with multiprocessing.Pool(
                self.core_cnt, maxtasksperchild=self.config.maxtasksperchild
            ) as pool:
                constants = functools.partial(
                    pose_plotter_mp,
                    video_meta_data=video_meta_data,
                    video_path=video_path,
                    bp_dict=self.config.animal_bp_dict,
                    circle_size=self.circle_size,
                    video_save_dir=self.temp_folder,
                )
                for cnt, result in enumerate(
                    pool.imap(
                        constants,
                        pose_lst,
                        chunksize=self.config.multiprocess_chunksize,
                    )
                ):
                    print(
                        f"Image {obs_per_split*(cnt+1)}/{len(pose_df)}, Video {file_cnt+1}/{len(list(self.data.keys()))}..."
                    )
                pool.terminate()
                pool.join()

            print(f"Joining {video_name} multi-processed video...")
            concatenate_videos_in_folder(
                in_folder=self.temp_folder,
                save_path=save_video_path,
                remove_splits=True,
            )
            video_timer.stop_timer()
            stdout_success(
                msg=f"Pose video {video_name} complete",
                elapsed_time=video_timer.elapsed_time_str,
                source=self.__class__.__name__,
            )
        self.config.timer.stop_timer()
        stdout_success(
            f"Pose visualizations for {len(list(self.data.keys()))} video(s) created in {self.out_dir} directory",
            elapsed_time=self.config.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# test = PosePlotter(in_dir='/Users/simon/Desktop/envs/troubleshooting/Zelikowsky/project_folder/csv/input_csv',
#                    out_dir='/Users/simon/Desktop/envs/troubleshooting/Zelikowsky/project_folder/test_viz',
#                    circle_size=10,
#                    core_cnt=1,
#                    color_settings={'Animal_1':  'Green', 'Animal_2':  'Red'})
# test.run()


# test = PosePlotter(in_dir='/Users/simon/Desktop/envs/troubleshooting/dam_nest-c-only_ryan/project_folder/csv/outlier_corrected_movement_location',
#                    out_dir='/Users/simon/Desktop/video_tests_',
#                    sample_time=2,
#                    circle_size=10,
#                    core_cnt=1,
#                    color_settings=None) #
# test.run()


# test = PosePlotter(in_dir='/Users/simon/Desktop/envs/troubleshooting/piotr/project_folder/csv/outlier_corrected_movement_location',
#                    out_dir='/Users/simon/Desktop/envs/troubleshooting/piotr/project_folder/frames/output/test',
#                    circle_size=10,
#                    core_cnt=6,
#                    color_settings={'Animal_1':  'Green', 'Animal_2':  'Red'})
# test.run()



__author__ = "Simon Nilsson"

import itertools
import os
from typing import Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_that_dir_has_list_of_filenames)
from simba.utils.enums import TagNames
from simba.utils.errors import AnimalNumberError, CountError, InvalidInputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class DirectingOtherAnimalsAnalyzer(ConfigReader, FeatureExtractionMixin):
    """
    Calculate when animals are directing towards body-parts of other animals. Results are stored in
    the ``project_folder/logs/directionality_dataframes`` directory of the SimBA project.

    .. note:
       `Example expected bool table <https://github.com/sgoldenlab/simba/blob/master/misc/boolean_directionaly_example.csv>`__.
       `Example expected summary table <https://github.com/sgoldenlab/simba/blob/master/misc/detailed_summary_directionality_example.csv>`__.
       `Example expected aggregate statistics table <https://github.com/sgoldenlab/simba/blob/master/misc/direction_data_aggregates_example.csv>`__.

    .. important::
       Requires the pose-estimation data for the ``left ear``, ``right ear`` and ``nose`` of each individual animals.
       `Github Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-3-generating-features-from-roi-data>`__.
       `Expected output <https://github.com/sgoldenlab/simba/blob/master/misc/Direction_data_example.csv>`__.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter bool bool_tables: If True, creates boolean output tables.
    :parameter bool summary_tables: If True, creates summary tables including approximate location of eye of observer and the location of observed body-parts and frames when observation was detected.
    :parameter bool aggregate_statistics_tables: If True, summary statistics tables of how much time each animal spent observation the other animals.

    :examples:
    >>> directing_analyzer = DirectingOtherAnimalsAnalyzer(config_path='MyProjectConfig')
    >>> directing_analyzer.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        bool_tables: Optional[bool] = True,
        summary_tables: Optional[bool] = False,
        append_bool_tables_to_features: Optional[bool] = False,
        aggregate_statistics_tables: Optional[bool] = False,
    ):

        super().__init__(config_path=config_path)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        if self.animal_cnt < 2:
            raise AnimalNumberError(
                "Cannot analyze directionality between animals in a 1 animal project.",
                source=self.__class__.__name__,
            )
        check_if_filepath_list_is_empty(
            filepaths=self.outlier_corrected_paths,
            error_msg=f"SIMBA ERROR: No data found in the {self.outlier_corrected_dir} directory",
        )
        self.animal_permutations = list(itertools.permutations(self.animal_bp_dict, 2))
        (
            self.bool_tables,
            self.summary_tables,
            self.aggregate_statistics_tables,
            self.append_bool_tables_to_features,
        ) = (
            bool_tables,
            summary_tables,
            aggregate_statistics_tables,
            append_bool_tables_to_features,
        )
        if self.append_bool_tables_to_features:
            check_that_dir_has_list_of_filenames(
                dir=self.features_dir,
                file_name_lst=self.outlier_corrected_paths,
                file_type=self.file_type,
            )
        print(f"Processing {str(len(self.outlier_corrected_paths))} video(s)...")

    def run(self):
        self.results_dict = {}
        for file_cnt, file_path in enumerate(self.outlier_corrected_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(file_path)
            print(f"Analyzing directionality between animals in video {video_name}...")
            self.results_dict[video_name] = {}
            data_df = read_df(file_path, self.file_type)
            direct_bp_dict = self.check_directionality_cords()
            for animal_permutation in self.animal_permutations:
                self.results_dict[video_name][
                    "{} {} {}".format(
                        animal_permutation[0],
                        "directing towards",
                        animal_permutation[1],
                    )
                ] = {}
                first_animal_bps, second_animal_bps = (
                    direct_bp_dict[animal_permutation[0]],
                    self.animal_bp_dict[animal_permutation[1]],
                )
                first_ear_left_arr = data_df[
                    [
                        first_animal_bps["Ear_left"]["X_bps"],
                        first_animal_bps["Ear_left"]["Y_bps"],
                    ]
                ].to_numpy()
                first_ear_right_arr = data_df[
                    [
                        first_animal_bps["Ear_right"]["X_bps"],
                        first_animal_bps["Ear_right"]["Y_bps"],
                    ]
                ].to_numpy()
                first_nose_arr = data_df[
                    [
                        first_animal_bps["Nose"]["X_bps"],
                        first_animal_bps["Nose"]["Y_bps"],
                    ]
                ].to_numpy()
                other_animal_x_bps, other_animal_y_bps = (
                    second_animal_bps["X_bps"],
                    second_animal_bps["Y_bps"],
                )
                for x_bp, y_bp in zip(other_animal_x_bps, other_animal_y_bps):
                    target_cord_arr = data_df[[x_bp, y_bp]].to_numpy()
                    direction_data = self.jitted_line_crosses_to_nonstatic_targets(
                        left_ear_array=first_ear_left_arr,
                        right_ear_array=first_ear_right_arr,
                        nose_array=first_nose_arr,
                        target_array=target_cord_arr,
                    )
                    x_min = np.minimum(direction_data[:, 1], first_nose_arr[:, 0])
                    y_min = np.minimum(direction_data[:, 2], first_nose_arr[:, 1])
                    delta_x = abs((direction_data[:, 1] - first_nose_arr[:, 0]) / 2)
                    delta_y = abs((direction_data[:, 2] - first_nose_arr[:, 1]) / 2)
                    x_middle, y_middle = np.add(x_min, delta_x), np.add(y_min, delta_y)
                    direction_data = np.concatenate(
                        (y_middle.reshape(-1, 1), direction_data), axis=1
                    )
                    direction_data = np.concatenate(
                        (x_middle.reshape(-1, 1), direction_data), axis=1
                    )
                    direction_data = np.delete(direction_data, [2, 3, 4], 1)
                    direction_data = np.hstack((direction_data, target_cord_arr))
                    bp_data = pd.DataFrame(
                        direction_data,
                        columns=["Eye_x", "Eye_y", "Directing_BOOL", x_bp, y_bp],
                    )
                    bp_data = bp_data[["Eye_x", "Eye_y", x_bp, y_bp, "Directing_BOOL"]]
                    bp_data.insert(loc=0, column="Animal_2_body_part", value=x_bp[:-2])
                    bp_data.insert(
                        loc=0, column="Animal_2", value=animal_permutation[1]
                    )
                    bp_data.insert(
                        loc=0, column="Animal_1", value=animal_permutation[0]
                    )
                    self.results_dict[video_name][
                        "{} {} {}".format(
                            animal_permutation[0],
                            "directing towards",
                            animal_permutation[1],
                        )
                    ][x_bp[:-2]] = bp_data
            video_timer.stop_timer()
            print(
                f"Direction analysis complete for video {video_name} ({file_cnt + 1}/{len(self.outlier_corrected_paths)}, elapsed time: {video_timer.elapsed_time_str}s)..."
            )
        if self.bool_tables:
            self.create_bool_tables()
        if self.summary_tables:
            self._transpose_results_to_df()
            self._save_directionality_dfs()
        if self.aggregate_statistics_tables:
            self.summary_statistics()

    def create_bool_tables(self):
        save_dir = os.path.join(
            self.logs_path, f"Animal_directing_animal_booleans_{self.datetime}"
        )
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)
        for video_cnt, (video_name, video_data) in enumerate(self.results_dict.items()):
            video_df = pd.DataFrame()
            print(
                f"Saving boolean directing tables for video {video_name} (Video {video_cnt+1}/{len(self.results_dict.keys())})..."
            )
            for animal_permutation, animal_permutation_data in video_data.items():
                for body_part_name, body_part_data in animal_permutation_data.items():
                    video_df[f"{animal_permutation}_{body_part_name}"] = body_part_data[
                        "Directing_BOOL"
                    ]
            if self.append_bool_tables_to_features:
                print(
                    f"Adding directionality tables to features data for video {video_name}..."
                )
                df = read_df(
                    file_path=os.path.join(
                        self.features_dir, f"{video_name}.{self.file_type}"
                    ),
                    file_type=self.file_type,
                )
                if len(df) != len(video_df):
                    raise CountError(
                        msg=f"Failed to join data files as they contains different number of frames: the file representing video {video_name} in directory {self.outlier_corrected_dir} contains {len(video_df)} frames, and the file representing video {video_name} in directory {self.features_dir} contains {len(df)} frames."
                    )
                else:
                    df = pd.concat(
                        [df.reset_index(drop=True), video_df.reset_index(drop=True)],
                        axis=1,
                    )
                    write_df(
                        df=df,
                        file_type=self.file_type,
                        save_path=os.path.join(
                            self.features_dir, f"{video_name}.{self.file_type}"
                        ),
                    )
            video_df.to_csv(os.path.join(save_dir, f"{video_name}.csv"))
        stdout_success(
            msg=f"All boolean tables saved in {save_dir}!",
            source=self.__class__.__name__,
        )

    def _transpose_results_to_df(self):
        """
        Privet method to transpose results created by :meth:`~simba.DirectingOtherAnimalsAnalyzer.process_directionality`.
        into dict of dataframes.
        """

        print("Transposing directionality data for summary tables...")
        self.directionality_df_dict = {}
        for video_name, video_data in self.results_dict.items():
            out_df_lst = []
            for animal_permutation, permutation_data in video_data.items():
                for bp_name, bp_data in permutation_data.items():
                    directing_df = (
                        bp_data[bp_data["Directing_BOOL"] == 1]
                        .reset_index()
                        .rename(
                            columns={
                                "index": "Frame_#",
                                bp_name + "_x": "Animal_2_bodypart_x",
                                bp_name + "_y": "Animal_2_bodypart_y",
                            }
                        )
                    )
                    directing_df.insert(loc=0, column="Video", value=video_name)
                    out_df_lst.append(directing_df)
            self.directionality_df_dict[video_name] = pd.concat(
                out_df_lst, axis=0
            ).drop("Directing_BOOL", axis=1)

    def _save_directionality_dfs(self):
        """
        Privat method to save result created by :meth:`~simba.DirectingOtherAnimalsAnalyzer.create_directionality_dfs`.
        into CSV files on disk. Results are stored in `project_folder/logs` directory of the SimBA project.
        """
        save_dir = os.path.join(
            self.logs_path,
            f"detailed_directionality_summary_dataframes_{self.datetime}",
        )
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        for video_cnt, (video_name, video_data) in enumerate(
            self.directionality_df_dict.items()
        ):
            save_name = os.path.join(save_dir, f"{video_name}.csv")
            video_data.to_csv(save_name)
            print(
                f"Detailed directional summary tables saved for video {video_name} (Video {video_cnt+1}/{len(list(self.directionality_df_dict.keys()))})..."
            )
        stdout_success(
            f"All detailed directional data saved in the {save_dir} directory!",
            source=self.__class__.__name__,
        )

    def summary_statistics(self):
        """
        Method to save aggregate statistics of data created by :meth:`~simba.DirectingOtherAnimalsAnalyzer.create_directionality_dfs`.
        into CSV files on disk. Results are stored in `project_folder/logs` directory of the SimBA project.
        """
        print("Computing summary statistics...")
        out_df_lst = []
        for video_name, video_data in self.results_dict.items():
            _, _, fps = self.read_video_info(video_name=video_name)
            for animal_permutation, permutation_data in video_data.items():
                idx_directing = set()
                for bp_name, bp_data in permutation_data.items():
                    idx_directing.update(
                        list(bp_data.index[bp_data["Directing_BOOL"] == 1])
                    )
                value = round(len(idx_directing) / fps, 3)
                out_df_lst.append(
                    pd.DataFrame(
                        [[video_name, animal_permutation, value]],
                        columns=["Video", "Animal permutation", "Value (s)"],
                    )
                )
        self.summary_df = (
            pd.concat(out_df_lst, axis=0)
            .sort_values(by=["Video", "Animal permutation"])
            .set_index("Video")
        )
        self.save_path = os.path.join(
            self.logs_path, f"Direction_aggregate_summary_data_{self.datetime}.csv"
        )
        self.summary_df.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Summary directional statistics saved at {self.save_path}",
            source=self.__class__.__name__,
        )
        stdout_success(
            msg="All directional data saved in SimBA project",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# test = DirectingOtherAnimalsAnalyzer(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                      bool_tables=True,
#                                      summary_tables=False,
#                                      aggregate_statistics_tables=False,
#                                      append_bool_tables_to_features=True)
# test.run()



__author__ = "Simon Nilsson"

import tkinter.ttk as ttk
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (DropDownMenu, Entry_Box, FileSelect,
                                        hxtScrollbar)
from simba.unsupervised.dataset_creator import DatasetCreator
from simba.unsupervised.enums import UMLOptions, Unsupervised
from simba.unsupervised.pop_up_classes import (
    ClusterFrequentistStatisticsPopUp, ClusterVisualizerPopUp, ClusterXAIPopUp,
    DataExtractorPopUp, DBCVPopUp, EmbedderCorrelationsPopUp,
    FitClusterModelsPopUp, FitDimReductionPopUp, GridSearchVisualizerPopUp,
    PrintEmBeddingInfoPopUp, TransformClustererPopUp,
    TransformDimReductionPopUp)
from simba.utils.enums import Formats


class UnsupervisedGUI(ConfigReader, PopUpMixin):
    """
    Main access to unsupervised interface
    """

    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        PopUpMixin.__init__(
            self,
            title="UNSUPERVISED ANALYSIS",
            config_path=config_path,
            size=(1000, 800),
        )
        self.main_frm = Toplevel()
        self.main_frm.minsize(1000, 800)
        self.main_frm.wm_title("UNSUPERVISED ANALYSIS")
        self.main_frm.columnconfigure(0, weight=1)
        self.main_frm.rowconfigure(0, weight=1)
        self.main_frm = ttk.Notebook(hxtScrollbar(self.main_frm))
        self.create_dataset_tab = ttk.Frame(self.main_frm)
        self.dimensionality_reduction_tab = ttk.Frame(self.main_frm)
        self.clustering_tab = ttk.Frame(self.main_frm)
        self.visualization_tab = ttk.Frame(self.main_frm)
        self.metrics_tab = ttk.Frame(self.main_frm)

        self.main_frm.add(
            self.create_dataset_tab,
            text=f'{"[CREATE DATASET]": ^20s}',
            compound="left",
            image=self.menu_icons["features"]["img"],
        )
        self.main_frm.add(
            self.dimensionality_reduction_tab,
            text=f'{"[DIMENSIONALITY REDUCTION]": ^20s}',
            compound="left",
            image=self.menu_icons["dimensionality_reduction"]["img"],
        )
        self.main_frm.add(
            self.clustering_tab,
            text=f'{"[CLUSTERING]": ^20s}',
            compound="left",
            image=self.menu_icons["cluster"]["img"],
        )
        self.main_frm.add(
            self.visualization_tab,
            text=f'{"[VISUALIZATION]": ^20s}',
            compound="left",
            image=self.menu_icons["visualize"]["img"],
        )
        self.main_frm.add(
            self.metrics_tab,
            text=f'{"[METRICS]": ^20s}',
            compound="left",
            image=self.menu_icons["metrics"]["img"],
        )
        self.main_frm.grid(row=0)

        self.clf_slice_options = [f"ALL CLASSIFIERS ({len(self.clf_names)})"]
        for clf_name in self.clf_names:
            self.clf_slice_options.append(f"{clf_name}")
        create_dataset_frm = LabelFrame(
            self.create_dataset_tab,
            text="CREATE DATASET",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.feature_file_selected = FileSelect(
            create_dataset_frm, "FEATURE FILE (CSV)", lblwidth=25
        )

        self.data_slice_dropdown = DropDownMenu(
            create_dataset_frm,
            "FEATURE SLICE:",
            UMLOptions.FEATURE_SLICE_OPTIONS.value,
            "25",
            com=lambda x: self.change_status_of_file_select(),
        )
        self.data_slice_dropdown.setChoices(UMLOptions.FEATURE_SLICE_OPTIONS.value[0])
        self.clf_slice_dropdown = DropDownMenu(
            create_dataset_frm, "CLASSIFIER SLICE:", self.clf_slice_options, "25"
        )
        self.clf_slice_dropdown.setChoices(self.clf_slice_options[0])
        self.change_status_of_file_select()
        self.bout_dropdown = DropDownMenu(
            create_dataset_frm,
            "BOUT AGGREGATION METHOD:",
            UMLOptions.BOUT_AGGREGATION_METHODS.value,
            "25",
        )
        self.bout_dropdown.setChoices(
            choice=UMLOptions.BOUT_AGGREGATION_METHODS.value[0]
        )
        self.min_bout_length = Entry_Box(
            create_dataset_frm, "MINIMUM BOUT LENGTH (MS): ", "25", validation="numeric"
        )
        self.min_bout_length.entry_set(val=0)
        self.create_btn = Button(
            create_dataset_frm,
            text="CREATE DATASET",
            fg="blue",
            command=lambda: self.create_dataset(),
        )

        create_dataset_frm.grid(row=0, column=0, sticky=NW)
        self.data_slice_dropdown.grid(row=0, column=0, sticky=NW)
        self.clf_slice_dropdown.grid(row=1, column=0, sticky=NW)
        self.bout_dropdown.grid(row=2, column=0, sticky=NW)
        self.feature_file_selected.grid(row=3, column=0, sticky=NW)
        self.min_bout_length.grid(row=4, column=0, sticky=NW)
        self.create_btn.grid(row=5, column=0, sticky=NW)

        self.dim_reduction_frm = LabelFrame(
            self.dimensionality_reduction_tab,
            text="DIMENSIONALITY REDUCTION",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.dim_reduction_fit_btn = Button(
            self.dim_reduction_frm,
            text="DIMENSIONALITY REDUCTION: FIT",
            fg="blue",
            command=lambda: FitDimReductionPopUp(),
        )
        self.dim_reduction_transform_btn = Button(
            self.dim_reduction_frm,
            text="DIMENSIONALITY REDUCTION: TRANSFORM",
            fg="red",
            command=lambda: TransformDimReductionPopUp(),
        )
        self.dim_reduction_frm.grid(row=0, column=0, sticky=NW)
        self.dim_reduction_fit_btn.grid(row=1, column=0, sticky=NW)
        self.dim_reduction_transform_btn.grid(row=2, column=0, sticky=NW)

        self.clustering_frm = LabelFrame(
            self.clustering_tab,
            text="CLUSTERING",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.cluster_fit_btn = Button(
            self.clustering_frm,
            text="CLUSTERING: FIT ",
            fg="blue",
            command=lambda: FitClusterModelsPopUp(),
        )
        self.cluster_transform_btn = Button(
            self.clustering_frm,
            text="CLUSTERING: TRANSFORM ",
            fg="red",
            command=lambda: TransformClustererPopUp(),
        )
        self.clustering_frm.grid(row=0, column=0, sticky=NW)
        self.cluster_fit_btn.grid(row=1, column=0, sticky=NW)
        self.cluster_transform_btn.grid(row=2, column=0, sticky=NW)

        self.visualization_frm = LabelFrame(
            self.visualization_tab,
            text="VISUALIZATIONS",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.grid_search_visualization_btn = Button(
            self.visualization_frm,
            text="GRID-SEARCH VISUALIZATION",
            fg="blue",
            command=lambda: GridSearchVisualizerPopUp(config_path=self.config_path),
        )
        self.cluster_visualizer = Button(
            self.visualization_frm,
            text="DATA VISUALIZERS",
            fg="red",
            command=lambda: ClusterVisualizerPopUp(config_path=self.config_path),
        )
        self.visualization_frm.grid(row=0, column=0, sticky="NW")
        self.grid_search_visualization_btn.grid(row=0, column=0, sticky="NW")
        self.cluster_visualizer.grid(row=1, column=0, sticky="NW")

        self.metrics_frm = LabelFrame(
            self.metrics_tab,
            text="METRICS",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.dbcv_btn = Button(
            self.metrics_frm,
            text="DENSITY-BASED CLUSTER VALIDATION",
            fg="blue",
            command=lambda: DBCVPopUp(config_path=self.config_path),
        )
        self.extract_single_metrics_btn = Button(
            self.metrics_frm,
            text="EXTRACT DATA",
            fg="red",
            command=lambda: DataExtractorPopUp(config_path=self.config_path),
        )
        self.cluster_descriptives_btn = Button(
            self.metrics_frm,
            text="CLUSTER FREQUENTIST STATISTICS",
            fg="green",
            command=lambda: ClusterFrequentistStatisticsPopUp(
                config_path=self.config_path
            ),
        )
        self.cluster_xai_btn = Button(
            self.metrics_frm,
            text="CLUSTER XAI STATISTICS",
            fg="blue",
            command=lambda: ClusterXAIPopUp(config_path=self.config_path),
        )
        self.embedding_corr_btn = Button(
            self.metrics_frm,
            text="EMBEDDING CORRELATIONS",
            fg="orange",
            command=lambda: EmbedderCorrelationsPopUp(config_path=self.config_path),
        )
        self.print_embedding_info_btn = Button(
            self.metrics_frm,
            text="PRINT MODEL INFO",
            fg="black",
            command=lambda: PrintEmBeddingInfoPopUp(config_path=self.config_path),
        )

        self.metrics_frm.grid(row=0, column=0, sticky="NW")
        self.dbcv_btn.grid(row=0, column=0, sticky="NW")
        self.extract_single_metrics_btn.grid(row=1, column=0, sticky="NW")
        self.cluster_descriptives_btn.grid(row=2, column=0, sticky="NW")
        self.cluster_xai_btn.grid(row=3, column=0, sticky="NW")
        self.embedding_corr_btn.grid(row=4, column=0, sticky="NW")
        self.print_embedding_info_btn.grid(row=5, column=0, sticky="NW")

        # self.main_frm.mainloop()

    def change_status_of_file_select(self):
        if self.data_slice_dropdown.getChoices() == "USER-DEFINED FEATURE SET":
            self.feature_file_selected.set_state(setstatus=NORMAL)
        else:
            self.feature_file_selected.set_state(setstatus=DISABLED)

    def create_dataset(self):
        data_slice_type = self.data_slice_dropdown.getChoices()
        classifier_slice_type = self.clf_slice_dropdown.getChoices()
        bout_selection = self.bout_dropdown.getChoices()
        bout_length = self.min_bout_length.entry_get
        feature_file_path = None
        if data_slice_type is Unsupervised.USER_DEFINED_SET.value:
            feature_file_path = self.feature_file_selected.file_path

        settings = {
            "data_slice": data_slice_type,
            "clf_slice": classifier_slice_type,
            "bout_aggregation_type": bout_selection,
            "min_bout_length": bout_length,
            "feature_file_path": feature_file_path,
        }
        _ = DatasetCreator(settings=settings, config_path=self.config_path)



__author__ = "Simon Nilsson"

from tkinter import *

import cv2
import numpy as np
from PIL import Image as Img
from PIL import ImageTk

from simba.utils.errors import FrameRangeError
from simba.utils.read_write import get_video_meta_data

PADDING = 5
MAX_SIZE = (1080, 650)


class InteractiveVideoPlotterWindow(object):
    def __init__(self, video_path: str, p_arr: np.array):
        self.main_frm = Toplevel()
        self.current_frm_number, self.jump_size = 0, 0
        self.img_frm = Frame(self.main_frm)
        self.img_frm.grid(row=0, column=1, sticky=NW)
        self.button_frame = Frame(self.main_frm, bd=2, width=700, height=300)
        self.button_frame.grid(row=1, column=0)
        self.video_meta_data = get_video_meta_data(video_path=video_path)
        self.cap = cv2.VideoCapture(video_path)
        self.cap.set(1, self.current_frm_number)
        self.max_frm = np.argmax(p_arr)
        self.frame_id_lbl = Label(self.button_frame, text="FRAME NUMBER")
        self.frame_id_lbl.grid(row=0, column=1, sticky=NW, padx=PADDING)
        self.forward_next_frm_btn = Button(
            self.button_frame,
            text=">",
            command=lambda: self.load_new_frame(frm_cnt=self.current_frm_number + 1),
        )
        self.forward_next_frm_btn.grid(row=1, column=3, sticky=E, padx=PADDING)

        self.forward_last_frm_btn = Button(
            self.button_frame,
            text=">>",
            command=lambda: self.load_new_frame(
                frm_cnt=self.video_meta_data["frame_count"] - 1
            ),
        )
        self.forward_last_frm_btn.grid(row=1, column=4, sticky=E, padx=PADDING)
        self.back_one_frm_btn = Button(
            self.button_frame,
            text="<",
            command=lambda: self.load_new_frame(frm_cnt=self.current_frm_number - 1),
        )
        self.back_one_frm_btn.grid(row=1, column=1, sticky=W, padx=PADDING)
        self.back_first_frm = Button(
            self.button_frame, text="<<", command=lambda: self.load_new_frame(frm_cnt=0)
        )
        self.back_first_frm.grid(row=1, column=0, sticky=W)
        self.frame_entry_var = IntVar(self.main_frm, value=self.current_frm_number)
        self.frame_entry_box = Entry(
            self.button_frame, width=7, textvariable=self.frame_entry_var
        )
        self.frame_entry_box.grid(row=1, column=1)
        self.select_frm_btn = Button(
            self.button_frame,
            text="Jump to selected frame",
            command=lambda: self.load_new_frame(frm_cnt=self.frame_entry_var.get()),
        )
        self.select_frm_btn.grid(row=2, column=1, sticky=N)

        self.jump_frm = Frame(self.main_frm)
        self.jump_frm.grid(row=2, column=0)
        self.jump_lbl = Label(self.jump_frm, text="Jump Size:")
        self.jump_lbl.grid(row=0, column=0, sticky=NW)
        self.jump_size_scale = Scale(
            self.jump_frm, from_=0, to=100, orient=HORIZONTAL, length=200
        )
        self.jump_size_scale.set(self.jump_size)
        self.jump_size_scale.grid(row=0, column=1, sticky=NW)
        self.jump_back_btn = Button(
            self.jump_frm,
            text="<<",
            command=lambda: self.load_new_frame(
                frm_cnt=self.current_frm_number - self.jump_size_scale.get()
            ),
        )
        self.jump_back_btn.grid(row=0, column=2, sticky=E)
        self.jump_forward_btn = Button(
            self.jump_frm,
            text=">>",
            command=lambda: self.load_new_frame(
                frm_cnt=self.current_frm_number + self.jump_size_scale.get()
            ),
        )
        self.jump_forward_btn.grid(row=0, column=3, sticky=W)

        self.load_new_frame(frm_cnt=self.current_frm_number)

        instructions_frm = Frame(self.main_frm, width=100, height=100)
        instructions_frm.grid(row=0, column=2, sticky=N)
        key_presses = Label(
            instructions_frm,
            text="\n\n Keyboard shortcuts for frame navigation: \n Right Arrow = +1 frame"
            "\n Left Arrow = -1 frame"
            "\n Ctrl + l = Last frame"
            "\n Ctrl + o = First frame",
        )

        move_to_highest_p_btn = Button(
            instructions_frm,
            text="SHOW HIGHEST \n PROBABILITY FRAME",
            command=lambda: self.load_new_frame(frm_cnt=self.max_frm),
        )
        key_presses.grid(row=0, column=0, sticky=S)
        move_to_highest_p_btn.grid(row=1, column=0, sticky=S)
        self.bind_keys()

    def bind_keys(self):
        self.main_frm.bind(
            "<Right>",
            lambda x: self.load_new_frame(frm_cnt=self.current_frm_number + 1),
        )
        self.main_frm.bind(
            "<Left>", lambda x: self.load_new_frame(frm_cnt=self.current_frm_number - 1)
        )
        self.main_frm.bind(
            "<Control-l>",
            lambda x: self.load_new_frame(
                frm_cnt=self.video_meta_data["frame_count"] - 1
            ),
        )
        self.main_frm.bind("<Control-o>", lambda x: self.load_new_frame(frm_cnt=0))

    def load_new_frame(self, frm_cnt: int):
        if (frm_cnt > self.video_meta_data["frame_count"] - 1) or (frm_cnt < 0):
            raise FrameRangeError(
                msg=f'Frame {str(frm_cnt)} is outside of the video frame range: (0-{self.video_meta_data["frame_count"]-1}).'
            )
        self.cap.set(1, int(frm_cnt))
        _, self.new_frm = self.cap.read()
        self.new_frm = cv2.cvtColor(self.new_frm, cv2.COLOR_RGB2BGR)
        self.new_frm = Img.fromarray(self.new_frm)
        self.new_frm.thumbnail(MAX_SIZE, Img.ANTIALIAS)
        self.new_frm = ImageTk.PhotoImage(master=self.main_frm, image=self.new_frm)
        self.img_frm = Label(self.main_frm, image=self.new_frm)
        self.img_frm.image = self.new_frm
        self.img_frm.grid(row=0, column=0)
        self.current_frm_number = frm_cnt
        self.frame_entry_var.set(value=self.current_frm_number)



__author__ = "Simon Nilsson"

import os
from datetime import datetime
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
from numba import jit

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df
from simba.utils.warnings import NoDataFoundWarning


class SeverityCalculator(ConfigReader):
    """
    Computes the "severity" of classification frame events based on how much
    the animals are moving. Frames are scored as less or more severe at lower and higher movements, respectively.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter dict settings: how to calculate the severity. E.g., {'brackets': 10, 'clf': 'Attack', 'animals': ['Simon', 'JJ'], 'time': True, 'frames': False}.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md>`__.

    Examples
    ----------
    >>> settings = {'brackets': 10, 'clf': 'Attack', 'animals': ['Simon', 'JJ'], 'time': True, 'frames': False}
    >>> processor = SeverityCalculator(config_path='project_folder/project_config.ini', settings=settings)
    >>> processor.run()
    >>> processor.save()
    """

    def __init__(self, config_path: Union[str, os.PathLike], settings: Dict):
        ConfigReader.__init__(self, config_path=config_path)
        self.settings = settings
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg=f"SIMBA ERROR: Cannot process severity. {self.machine_results_dir} directory is empty",
        )
        save_name = os.path.join(
            f'severity_{datetime.now().strftime("%Y%m%d%H%M%S")}.csv'
        )
        self.save_path = os.path.join(self.logs_path, save_name)
        self.results = {}

    @staticmethod
    @jit(nopython=True)
    def __euclidean_distance(bp_1_x_vals, bp_2_x_vals, bp_1_y_vals, bp_2_y_vals):
        return np.sqrt(
            (bp_1_x_vals - bp_2_x_vals) ** 2 + (bp_1_y_vals - bp_2_y_vals) ** 2
        )

    def run(self):
        for file_path in self.machine_results_paths:
            _, video_name, _ = get_fn_ext(file_path)
            self.results[video_name] = {}
            df = read_df(file_path=file_path, file_type=self.file_type)
            if self.settings["clf"] not in df.columns:
                NoDataFoundWarning(
                    msg=f'Skipping file {video_name} - {self.settings["clf"]} data not present in file'
                )
                continue
            _, _, fps = self.read_video_info(video_name=video_name)
            for animal_name, animal_bodyparts in self.animal_bp_dict.items():
                animal_df = df[animal_bodyparts["X_bps"] + animal_bodyparts["Y_bps"]]
                shifted = animal_df.shift(periods=1).fillna(0)
                movement = pd.DataFrame()
                for bp_x, bp_y in zip(
                    animal_bodyparts["X_bps"], animal_bodyparts["Y_bps"]
                ):
                    movement[bp_x.rstrip("_x")] = self.__euclidean_distance(
                        animal_df[bp_x].values,
                        shifted[bp_x].values,
                        animal_df[bp_y].values,
                        shifted[bp_y].values,
                    )
                movement["sum"] = movement.sum(axis=1)
                movement["sum"].iloc[0] = 0
                df[animal_name] = movement["sum"]
            df["movement"] = df[self.settings["animals"]].sum(axis=1)
            df["bin"] = pd.qcut(
                x=df["movement"],
                q=self.settings["brackets"],
                labels=list(range(1, self.settings["brackets"] + 1)),
            )
            clf_df = (
                df["bin"][df[self.settings["clf"]] == 1]
                .astype(int)
                .reset_index(drop=True)
            )
            for i in range(0, self.settings["brackets"]):
                if self.settings["frames"]:
                    self.results[video_name][f"Grade {str(i + 1)} (frames)"] = len(
                        clf_df[clf_df == i]
                    )
                if self.settings["time"]:
                    self.results[video_name][f"Grade {str(i + 1)} (s)"] = round(
                        (len(clf_df[clf_df == i]) / fps), 4
                    )

    def save(self):
        out_df = pd.DataFrame(columns=["VIDEO", "MEASUREMENT", "VALUE"])
        for video_name, video_data in self.results.items():
            for grade, grade_data in video_data.items():
                out_df.loc[len(out_df)] = [video_name, grade, grade_data]
        out_df.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Severity data saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# settings = {'brackets': 10, 'clf': 'Attack', 'animals': ['Simon', 'JJ'], 'time': True, 'frames': True}
# processor = SeverityCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', settings=settings)
# processor.run()
# processor.save()



# All credit to https://stackoverflow.com/questions/46571448/tkinter-and-a-html-file - thanks DELICA - https://stackoverflow.com/users/7027346/delica

import ctypes

from cefpython3 import cefpython as cef

try:
    import tkinter as tk
    from tkinter import messagebox
except ImportError:
    import Tkinter as tk

import logging as _logging
import platform
import sys

# Fix for PyCharm hints warnings
WindowUtils = cef.WindowUtils()

# Platforms
WINDOWS = platform.system() == "Windows"
LINUX = platform.system() == "Linux"
MAC = platform.system() == "Darwin"

# Globals
logger = _logging.getLogger("tkinter_.py")
url = "localhost:8050/"


class MainFrame(tk.Frame):
    def __init__(self, root):
        self.closing = False
        self.browser = None

        # Root
        root.geometry("900x640")
        tk.Grid.rowconfigure(root, 0, weight=1)
        tk.Grid.columnconfigure(root, 0, weight=1)

        # MainFrame
        tk.Frame.__init__(self, root)
        self.master.title("SimBA Dashboard")
        self.master.protocol("WM_DELETE_WINDOW", self.on_close)
        self.bind("<Configure>", self.on_configure)
        self.bind("<FocusIn>", self.on_focus_in)
        self.bind("<FocusOut>", self.on_focus_out)
        self.focus_set()

        # Pack MainFrame
        self.pack(fill=tk.BOTH, expand=tk.YES)

    def embed_browser(self):
        window_info = cef.WindowInfo()
        rect = [0, 0, self.winfo_width(), self.winfo_height()]
        window_info.SetAsChild(self.get_window_handle(), rect)
        self.browser = cef.CreateBrowserSync(window_info, url=url)  # todo
        assert self.browser
        self.browser.SetClientHandler(LoadHandler(self))
        self.browser.SetClientHandler(FocusHandler(self))
        self.message_loop_work()

    def get_window_handle(self):
        if self.winfo_id() > 0:
            return self.winfo_id()
        else:
            raise Exception("Couldn't obtain window handle")

    def message_loop_work(self):
        cef.MessageLoopWork()
        self.after(10, self.message_loop_work)

    def on_configure(self, event):
        width = event.width
        height = event.height
        if self.browser:
            if WINDOWS:
                ctypes.windll.user32.SetWindowPos(
                    self.browser.GetWindowHandle(), 0, 0, 0, width, height, 0x0002
                )
            elif LINUX:
                self.browser.SetBounds(0, 0, width, height)
            self.browser.NotifyMoveOrResizeStarted()
        if not self.browser:
            self.embed_browser()

    def on_focus_in(self, _):
        logger.debug("BrowserFrame.on_focus_in")
        if self.browser:
            self.browser.SetFocus(True)
            self.focus_set()

    def on_focus_out(self, _):
        logger.debug("BrowserFrame.on_focus_out")
        if self.browser:
            self.browser.SetFocus(False)

    def on_close(self):
        if self.browser:
            self.browser.CloseBrowser(True)
            self.clear_browser_references()
        self.destroy()
        self.master.destroy()

    def get_browser(self):
        if self.browser:
            return self.browser
        return None

    def clear_browser_references(self):
        self.browser = None


class LoadHandler(object):
    def __init__(self, browser_frame):
        self.browser_frame = browser_frame


class FocusHandler(object):
    def __init__(self, browser):
        self.browser = browser

    def OnTakeFocus(self, next_component, **_):
        logger.debug(
            "FocusHandler.OnTakeFocus, next={next}".format(next=next_component)
        )

    def OnSetFocus(self, source, **_):
        logger.debug("FocusHandler.OnSetFocus, source={source}".format(source=source))
        return False

    def OnGotFocus(self, **_):
        """Fix CEF focus issues (#255). Call browser frame's focus_set
        to get rid of type cursor in url entry widget."""
        logger.debug("FocusHandler.OnGotFocus")
        self.browser.focus_set()


# if __name__ == '__main__':
logger.setLevel(_logging.INFO)
stream_handler = _logging.StreamHandler()
formatter = _logging.Formatter("[%(filename)s] %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)
logger.info("CEF Python {ver}".format(ver=cef.__version__))
logger.info(
    "Python {ver} {arch}".format(
        ver=platform.python_version(), arch=platform.architecture()[0]
    )
)
logger.info("Tk {ver}".format(ver=tk.Tcl().eval("info patchlevel")))
assert cef.__version__ >= "55.3", "CEF Python v55.3+ required to run this"
sys.excepthook = cef.ExceptHook  # To shutdown all CEF processes on error
root = tk.Tk()
app = MainFrame(root)


def on_closing():
    if messagebox.askokcancel("Quit", "Do you want to quit?"):
        root.destroy()


root.protocol("WM_DELETE_WINDOW", on_closing)

# Tk must be initialized before CEF otherwise fatal error (Issue #306)
cef.Initialize()
root.mainloop()
# app.mainloop()
cef.Shutdown()



__author__ = "Simon Nilsson"

import itertools
import os

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.roi_tools.ROI_feature_analyzer import ROIFeatureCreator
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Formats
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data, read_df


class ROIfeatureVisualizer(ConfigReader):
    """
    Visualizing features that depend on the relationships between the location of the animals and user-defined
    ROIs. E.g., distances to centroids of ROIs, cumulative time spent in ROIs, if animals are directing towards ROIs
    etc.

    .. note::
       For improved run-time, see :meth:`simba.ROI_feature_visualizer_mp.ROIfeatureVisualizerMultiprocess` for multiprocess class.
       `Tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-5-visualizing-roi-features>`__.
        .. image:: _static/img/roi_visualize.png
           :width: 400
           :align: center



    :param str config_path: Path to SimBA project config file in Configparser format
    :param str video_name: Name of video to create feature visualizations for.
    :param dict style_attr: User-defined styles (sizes, colors etc.)

    :example:
    >>> style_attr = {'ROI_centers': True, 'ROI_ear_tags': True, 'Directionality': True, 'Border_color': (0, 128, 0), 'Pose_estimation': True}
    >>> _ = ROIfeatureVisualizer(config_path='MyProjectConfig', video_name='MyVideo.mp4', style_attr=style_attr).run()
    """

    def __init__(self, config_path: str, video_name: str, style_attr: dict):
        super().__init__(config_path=config_path)
        _, self.video_name, _ = get_fn_ext(video_name)
        self.save_path = os.path.join(
            self.roi_features_save_dir, self.video_name + ".mp4"
        )
        if not os.path.exists(self.roi_features_save_dir):
            os.makedirs(self.roi_features_save_dir)
        c_map_size, self.style_attr = int(len(self.x_cols) + 1), style_attr
        self.roi_feature_creator = ROIFeatureCreator(config_path=config_path)
        self.file_in_path = os.path.join(
            self.outlier_corrected_dir, self.video_name + "." + self.file_type
        )
        self.video_path = os.path.join(self.project_path, "videos", video_name)
        check_file_exist_and_readable(file_path=self.file_in_path)
        self.roi_feature_creator.features_files = [self.file_in_path]
        self.roi_feature_creator.files_found = [self.file_in_path]
        self.roi_feature_creator.run()
        self.video_meta_data = get_video_meta_data(self.video_path)
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.font = cv2.FONT_HERSHEY_COMPLEX
        self.cap = cv2.VideoCapture(self.video_path)
        self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
            25,
            10,
            1500,
            0.8,
        )
        self.max_dim = max(
            self.video_meta_data["width"], self.video_meta_data["height"]
        )
        self.circle_scale = int(self.radius_scale / (self.res_scale / self.max_dim))
        self.font_size = float(self.font_scale / (self.res_scale / self.max_dim))
        self.spacing_scale = int(self.space_scale / (self.res_scale / self.max_dim))
        self.data_df = read_df(self.file_in_path, self.file_type)
        self.bp_names = self.roi_feature_creator.roi_analyzer.bp_dict
        self.video_recs = self.roi_feature_creator.roi_analyzer.video_recs
        self.video_circs = self.roi_feature_creator.roi_analyzer.video_circs
        self.video_polys = self.roi_feature_creator.roi_analyzer.video_polys
        self.shape_dicts = {}
        for df in [self.video_recs, self.video_circs, self.video_polys]:
            d = df.set_index("Name").to_dict(orient="index")
            self.shape_dicts = {**self.shape_dicts, **d}
        self.video_shapes = list(
            itertools.chain(
                self.video_recs["Name"].unique(),
                self.video_circs["Name"].unique(),
                self.video_polys["Name"].unique(),
            )
        )
        self.roi_directing_viable = self.roi_feature_creator.roi_directing_viable
        self.directing_data = None
        if self.roi_directing_viable:
            self.directing_data = self.roi_feature_creator.directing_analyzer.results_df
            self.directing_data = self.directing_data[
                self.directing_data["Video"] == self.video_name
            ]
        self.roi_feature_creator.out_df.fillna(0, inplace=True)

    def __calc_text_locs(self):
        add_spacer = 2
        self.loc_dict = {}
        for animal_cnt, animal_name in enumerate(self.multi_animal_id_list):
            self.loc_dict[animal_name] = {}
            for shape in self.video_shapes:
                self.loc_dict[animal_name][shape] = {}
                self.loc_dict[animal_name][shape]["in_zone_text"] = "{} {} {}".format(
                    shape, animal_name, "in zone"
                )
                self.loc_dict[animal_name][shape]["distance_text"] = "{} {} {}".format(
                    shape, animal_name, "distance"
                )
                self.loc_dict[animal_name][shape]["in_zone_text_loc"] = (
                    (self.video_meta_data["width"] + 5),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.spacing_scale * add_spacer
                    ),
                )
                self.loc_dict[animal_name][shape]["in_zone_data_loc"] = (
                    int(self.img_w_border_w - (self.img_w_border_w / 8)),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.spacing_scale * add_spacer
                    ),
                )
                add_spacer += 1
                self.loc_dict[animal_name][shape]["distance_text_loc"] = (
                    (self.video_meta_data["width"] + 5),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.spacing_scale * add_spacer
                    ),
                )
                self.loc_dict[animal_name][shape]["distance_data_loc"] = (
                    int(self.img_w_border_w - (self.img_w_border_w / 8)),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.spacing_scale * add_spacer
                    ),
                )
                add_spacer += 1
                if self.roi_directing_viable and self.style_attr["Directionality"]:
                    self.loc_dict[animal_name][shape]["directing_text"] = (
                        "{} {} {}".format(shape, animal_name, "facing")
                    )
                    self.loc_dict[animal_name][shape]["directing_text_loc"] = (
                        (self.video_meta_data["width"] + 5),
                        (
                            self.video_meta_data["height"]
                            - (self.video_meta_data["height"] + 10)
                            + self.spacing_scale * add_spacer
                        ),
                    )
                    self.loc_dict[animal_name][shape]["directing_data_loc"] = (
                        int(self.img_w_border_w - (self.img_w_border_w / 8)),
                        (
                            self.video_meta_data["height"]
                            - (self.video_meta_data["height"] + 10)
                            + self.spacing_scale * add_spacer
                        ),
                    )
                    add_spacer += 1

    def __insert_texts(self, shape_df):
        for animal_name in self.multi_animal_id_list:
            for _, shape in shape_df.iterrows():
                shape_name, shape_color = shape["Name"], shape["Color BGR"]
                cv2.putText(
                    self.img_w_border,
                    self.loc_dict[animal_name][shape_name]["in_zone_text"],
                    self.loc_dict[animal_name][shape_name]["in_zone_text_loc"],
                    self.font,
                    self.font_size,
                    shape_color,
                    1,
                )
                cv2.putText(
                    self.img_w_border,
                    self.loc_dict[animal_name][shape_name]["distance_text"],
                    self.loc_dict[animal_name][shape_name]["distance_text_loc"],
                    self.font,
                    self.font_size,
                    shape_color,
                    1,
                )
                if self.roi_directing_viable:
                    cv2.putText(
                        self.img_w_border,
                        self.loc_dict[animal_name][shape_name]["directing_text"],
                        self.loc_dict[animal_name][shape_name]["directing_text_loc"],
                        self.font,
                        self.font_size,
                        shape_color,
                        1,
                    )

    def __insert_directing_line(self):
        r = self.directing_data.loc[
            (self.directing_data["ROI"] == self.shape_name)
            & (self.directing_data["Animal"] == self.animal_name)
            & (self.directing_data["Frame"] == self.frame_cnt)
        ]
        clr = self.shape_dicts[self.shape_name]["Color BGR"]
        thickness = self.shape_dicts[self.shape_name]["Thickness"]
        if self.style_attr["Directionality_style"] == "Funnel":
            convex_hull_arr = (
                np.array(
                    [
                        [r["ROI_edge_1_x"], r["ROI_edge_1_y"]],
                        [r["ROI_edge_2_x"], r["ROI_edge_2_y"]],
                        [r["Eye_x"], r["Eye_y"]],
                    ]
                )
                .reshape(-1, 2)
                .astype(int)
            )
            cv2.fillPoly(self.img_w_border, [convex_hull_arr], clr)

        if self.style_attr["Directionality_style"] == "Lines":
            cv2.line(
                self.img_w_border,
                (int(r["Eye_x"]), int(r["Eye_y"])),
                (int(r["ROI_x"]), int(r["ROI_y"])),
                clr,
                int(thickness),
            )

    def run(self):
        """
        Creates and saves visualizations of ROI-based features. Results are stored in the ``project_folder/frames/
        output/ROI_features`` directory  of the SimBA project.

        Returns
        ----------
        None
        """

        self.frame_cnt = 0
        while self.cap.isOpened():
            ret, self.frame = self.cap.read()
            try:
                if ret:
                    self.img_w_border = cv2.copyMakeBorder(
                        self.frame,
                        0,
                        0,
                        0,
                        int(self.video_meta_data["width"]),
                        borderType=cv2.BORDER_CONSTANT,
                        value=self.style_attr["Border_color"],
                    )
                    self.img_w_border_h, self.img_w_border_w = (
                        self.img_w_border.shape[0],
                        self.img_w_border.shape[1],
                    )
                    if self.frame_cnt == 0:
                        self.__calc_text_locs()
                        self.writer = cv2.VideoWriter(
                            self.save_path,
                            self.fourcc,
                            self.video_meta_data["fps"],
                            (self.img_w_border_w, self.img_w_border_h),
                        )
                    self.__insert_texts(self.video_recs)
                    self.__insert_texts(self.video_circs)
                    self.__insert_texts(self.video_polys)

                    if self.style_attr["Pose_estimation"]:
                        for animal, animal_bp_name in self.bp_names.items():
                            bp_cords = self.data_df.loc[
                                self.frame_cnt, animal_bp_name
                            ].values
                            cv2.circle(
                                self.img_w_border,
                                (int(bp_cords[0]), int(bp_cords[1])),
                                0,
                                self.animal_bp_dict[animal]["colors"][0],
                                self.circle_scale,
                            )
                            cv2.putText(
                                self.img_w_border,
                                animal,
                                (int(bp_cords[0]), int(bp_cords[1])),
                                self.font,
                                self.font_size,
                                self.animal_bp_dict[animal]["colors"][0],
                                1,
                            )

                    for _, row in self.video_recs.iterrows():
                        cv2.rectangle(
                            self.img_w_border,
                            (int(row["topLeftX"]), int(row["topLeftY"])),
                            (int(row["Bottom_right_X"]), int(row["Bottom_right_Y"])),
                            row["Color BGR"],
                            int(row["Thickness"]),
                        )
                        if self.style_attr["ROI_centers"]:
                            center_cord = (
                                (int(row["topLeftX"] + (row["width"] / 2))),
                                (int(row["topLeftY"] + (row["height"] / 2))),
                            )
                            cv2.circle(
                                self.img_w_border,
                                center_cord,
                                self.circle_scale,
                                row["Color BGR"],
                                -1,
                            )
                        if self.style_attr["ROI_centers"]:
                            for tag_data in row["Tags"].values():
                                cv2.circle(
                                    self.img_w_border,
                                    tuple(tag_data),
                                    self.circle_scale,
                                    row["Color BGR"],
                                    -1,
                                )

                    for _, row in self.video_circs.iterrows():
                        cv2.circle(
                            self.img_w_border,
                            (int(row["centerX"]), int(row["centerY"])),
                            row["radius"],
                            row["Color BGR"],
                            int(row["Thickness"]),
                        )
                        if self.style_attr["ROI_centers"]:
                            cv2.circle(
                                self.img_w_border,
                                (int(row["centerX"]), int(row["centerY"])),
                                self.circle_scale,
                                row["Color BGR"],
                                -1,
                            )
                        if self.style_attr["ROI_centers"]:
                            for tag_data in row["Tags"].values():
                                cv2.circle(
                                    self.img_w_border,
                                    tuple(tag_data),
                                    self.circle_scale,
                                    row["Color BGR"],
                                    -1,
                                )

                    for _, row in self.video_polys.iterrows():
                        cv2.polylines(
                            self.img_w_border,
                            [row["vertices"].astype(int)],
                            True,
                            row["Color BGR"],
                            thickness=int(row["Thickness"]),
                        )
                        if self.style_attr["ROI_centers"]:
                            cv2.circle(
                                self.img_w_border,
                                (int(row["Center_X"]), int(row["Center_Y"])),
                                self.circle_scale,
                                row["Color BGR"],
                                -1,
                            )
                        if self.style_attr["ROI_ear_tags"]:
                            for tag_data in row["vertices"]:
                                cv2.circle(
                                    self.img_w_border,
                                    tuple(tag_data),
                                    self.circle_scale,
                                    row["Color BGR"],
                                    -1,
                                )

                    for animal_name, shape_name in itertools.product(
                        self.multi_animal_id_list, self.video_shapes
                    ):
                        self.animal_name, self.shape_name = animal_name, shape_name
                        in_zone_col_name = "{} {} {}".format(
                            shape_name, animal_name, "in zone"
                        )
                        distance_col_name = "{} {} {}".format(
                            shape_name, animal_name, "distance"
                        )
                        in_zone_value = str(
                            bool(
                                self.roi_feature_creator.out_df.loc[
                                    self.frame_cnt, in_zone_col_name
                                ]
                            )
                        )
                        distance_value = round(
                            self.roi_feature_creator.out_df.loc[
                                self.frame_cnt, distance_col_name
                            ],
                            2,
                        )
                        cv2.putText(
                            self.img_w_border,
                            in_zone_value,
                            self.loc_dict[animal_name][shape_name]["in_zone_data_loc"],
                            self.font,
                            self.font_size,
                            self.shape_dicts[shape_name]["Color BGR"],
                            1,
                        )
                        cv2.putText(
                            self.img_w_border,
                            str(distance_value),
                            self.loc_dict[animal_name][shape_name]["distance_data_loc"],
                            self.font,
                            self.font_size,
                            self.shape_dicts[shape_name]["Color BGR"],
                            1,
                        )
                        if (
                            self.roi_directing_viable
                            and self.style_attr["Directionality"]
                        ):
                            facing_col_name = "{} {} {}".format(
                                shape_name, animal_name, "facing"
                            )
                            facing_value = self.roi_feature_creator.out_df.loc[
                                self.frame_cnt, facing_col_name
                            ]
                            cv2.putText(
                                self.img_w_border,
                                str(bool(facing_value)),
                                self.loc_dict[animal_name][shape_name][
                                    "directing_data_loc"
                                ],
                                self.font,
                                self.font_size,
                                self.shape_dicts[shape_name]["Color BGR"],
                                1,
                            )
                            if facing_value:
                                self.__insert_directing_line()

                    self.frame_cnt += 1
                    self.writer.write(np.uint8(self.img_w_border))
                    print(
                        "Frame: {} / {}. Video: {}".format(
                            str(self.frame_cnt),
                            str(self.video_meta_data["frame_count"]),
                            self.video_name,
                        )
                    )

                else:
                    self.timer.stop_timer()
                    self.cap.release()
                    self.writer.release()
                    print(
                        "Feature video {} saved in {} directory ...(elapsed time: {}s)".format(
                            self.video_name, self.save_path, self.timer.elapsed_time_str
                        )
                    )

            except:
                break

        self.timer.stop_timer()
        self.cap.release()
        self.writer.release()
        stdout_success(
            "Feature video {} saved in {} directory ...",
            elapsed_time=self.timer.elapsed_time_str,
        )


# style_attr = {'ROI_centers': True, 'ROI_ear_tags': True, 'Directionality': True, 'Border_color': (0, 128, 0), 'Pose_estimation': True}
# test = ROIfeatureVisualizer(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', video_name='Together_1.avi', style_attr=style_attr)
# test.run()


# style_attr = {'ROI_centers': True, 'ROI_ear_tags': True, 'Directionality': True, 'Directionality_style': 'Line', 'Border_color': (0, 128, 0), 'Pose_estimation': True}
# test = ROIfeatureVisualizer(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/project_config.ini', video_name='Video1.mp4', style_attr=style_attr)
# test.run()


# test = ROIfeatureVisualizer(config_path='/Users/simon/Desktop/train_model_project/project_folder/project_config.ini', video_name='Together_1.avi')
# test.run()
# test.save_new_features_files()
